#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict

from ..mixins.log import LogMixIn
from ..mixins.ondemand import OnDemandMixIn

GPG_KEY = """
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v2.0.22 (GNU/Linux)

mQENBE5OMmIBCAD+FPYKGriGGf7NqwKfWC83cBV01gabgVWQmZbMcFzeW+hMsgxH
W6iimD0RsfZ9oEbfJCPG0CRSZ7ppq5pKamYs2+EJ8Q2ysOFHHwpGrA2C8zyNAs4I
QxnZZIbETgcSwFtDun0XiqPwPZgyuXVm9PAbLZRbfBzm8wR/3SWygqZBBLdQk5TE
fDR+Eny/M1RVR4xClECONF9UBB2ejFdI1LD45APbP2hsN/piFByU1t7yK2gpFyRt
97WzGHn9MV5/TL7AmRPM4pcr3JacmtCnxXeCZ8nLqedoSuHFuhwyDnlAbu8I16O5
XRrfzhrHRJFM1JnIiGmzZi6zBvH0ItfyX6ttABEBAAG0KW5naW54IHNpZ25pbmcg
a2V5IDxzaWduaW5nLWtleUBuZ2lueC5jb20+iQE+BBMBAgAoAhsDBgsJCAcDAgYV
CAIJCgsEFgIDAQIeAQIXgAUCV2K1+AUJGB4fQQAKCRCr9b2Ce9m/YloaB/9XGrol
kocm7l/tsVjaBQCteXKuwsm4XhCuAQ6YAwA1L1UheGOG/aa2xJvrXE8X32tgcTjr
KoYoXWcdxaFjlXGTt6jV85qRguUzvMOxxSEM2Dn115etN9piPl0Zz+4rkx8+2vJG
F+eMlruPXg/zd88NvyLq5gGHEsFRBMVufYmHtNfcp4okC1klWiRIRSdp4QY1wdrN
1O+/oCTl8Bzy6hcHjLIq3aoumcLxMjtBoclc/5OTioLDwSDfVx7rWyfRhcBzVbwD
oe/PD08AoAA6fxXvWjSxy+dGhEaXoTHjkCbz/l6NxrK3JFyauDgU4K4MytsZ1HDi
MgMW8hZXxszoICTTiQEcBBABAgAGBQJOTkelAAoJEKZP1bF62zmo79oH/1XDb29S
YtWp+MTJTPFEwlWRiyRuDXy3wBd/BpwBRIWfWzMs1gnCjNjk0EVBVGa2grvy9Jtx
JKMd6l/PWXVucSt+U/+GO8rBkw14SdhqxaS2l14v6gyMeUrSbY3XfToGfwHC4sa/
Thn8X4jFaQ2XN5dAIzJGU1s5JA0tjEzUwCnmrKmyMlXZaoQVrmORGjCuH0I0aAFk
RS0UtnB9HPpxhGVbs24xXZQnZDNbUQeulFxS4uP3OLDBAeCHl+v4t/uotIad8v6J
SO93vc1evIje6lguE81HHmJn9noxPItvOvSMb2yPsE8mH4cJHRTFNSEhPW6ghmlf
Wa9ZwiVX5igxcvaIRgQQEQIABgUCTk5b0gAKCRDs8OkLLBcgg1G+AKCnacLb/+W6
cflirUIExgZdUJqoogCeNPVwXiHEIVqithAM1pdY/gcaQZmIRgQQEQIABgUCTk5f
YQAKCRCpN2E5pSTFPnNWAJ9gUozyiS+9jf2rJvqmJSeWuCgVRwCcCUFhXRCpQO2Y
Va3l3WuB+rgKjsQ=
=EWWI
-----END PGP PUBLIC KEY BLOCK-----

"""

MIME_TYPES = """
types {
    text/html                             html htm shtml;
    text/css                              css;
    text/xml                              xml;
    image/gif                             gif;
    image/jpeg                            jpeg jpg;
    application/javascript                js;
    application/atom+xml                  atom;
    application/rss+xml                   rss;

    text/mathml                           mml;
    text/plain                            txt;
    text/vnd.sun.j2me.app-descriptor      jad;
    text/vnd.wap.wml                      wml;
    text/x-component                      htc;

    image/png                             png;
    image/tiff                            tif tiff;
    image/vnd.wap.wbmp                    wbmp;
    image/x-icon                          ico;
    image/x-jng                           jng;
    image/x-ms-bmp                        bmp;
    image/svg+xml                         svg svgz;
    image/webp                            webp;

    application/font-woff                 woff;
    application/java-archive              jar war ear;
    application/json                      json;
    application/mac-binhex40              hqx;
    application/msword                    doc;
    application/pdf                       pdf;
    application/postscript                ps eps ai;
    application/rtf                       rtf;
    application/vnd.apple.mpegurl         m3u8;
    application/vnd.ms-excel              xls;
    application/vnd.ms-fontobject         eot;
    application/vnd.ms-powerpoint         ppt;
    application/vnd.wap.wmlc              wmlc;
    application/vnd.google-earth.kml+xml  kml;
    application/vnd.google-earth.kmz      kmz;
    application/x-7z-compressed           7z;
    application/x-cocoa                   cco;
    application/x-java-archive-diff       jardiff;
    application/x-java-jnlp-file          jnlp;
    application/x-makeself                run;
    application/x-perl                    pl pm;
    application/x-pilot                   prc pdb;
    application/x-rar-compressed          rar;
    application/x-redhat-package-manager  rpm;
    application/x-sea                     sea;
    application/x-shockwave-flash         swf;
    application/x-stuffit                 sit;
    application/x-tcl                     tcl tk;
    application/x-x509-ca-cert            der pem crt;
    application/x-xpinstall               xpi;
    application/xhtml+xml                 xhtml;
    application/xspf+xml                  xspf;
    application/zip                       zip;

    application/octet-stream              bin exe dll;
    application/octet-stream              deb;
    application/octet-stream              dmg;
    application/octet-stream              iso img;
    application/octet-stream              msi msp msm;

    application/vnd.openxmlformats-officedocument.wordprocessingml.document    docx;
    application/vnd.openxmlformats-officedocument.spreadsheetml.sheet          xlsx;
    application/vnd.openxmlformats-officedocument.presentationml.presentation  pptx;

    audio/midi                            mid midi kar;
    audio/mpeg                            mp3;
    audio/ogg                             ogg;
    audio/x-m4a                           m4a;
    audio/x-realaudio                     ra;

    video/3gpp                            3gpp 3gp;
    video/mp2t                            ts;
    video/mp4                             mp4;
    video/mpeg                            mpeg mpg;
    video/quicktime                       mov;
    video/webm                            webm;
    video/x-flv                           flv;
    video/x-m4v                           m4v;
    video/x-mng                           mng;
    video/x-ms-asf                        asx asf;
    video/x-ms-wmv                        wmv;
    video/x-msvideo                       avi;
}
"""

FASTCGI_PARAMS = """
fastcgi_param  QUERY_STRING       $query_string;
fastcgi_param  REQUEST_METHOD     $request_method;
fastcgi_param  CONTENT_TYPE       $content_type;
fastcgi_param  CONTENT_LENGTH     $content_length;

fastcgi_param  SCRIPT_NAME        $fastcgi_script_name;
fastcgi_param  REQUEST_URI        $request_uri;
fastcgi_param  DOCUMENT_URI       $document_uri;
fastcgi_param  DOCUMENT_ROOT      $document_root;
fastcgi_param  SERVER_PROTOCOL    $server_protocol;
fastcgi_param  REQUEST_SCHEME     $scheme;
fastcgi_param  HTTPS              $https if_not_empty;

fastcgi_param  GATEWAY_INTERFACE  CGI/1.1;
fastcgi_param  SERVER_SOFTWARE    nginx/$nginx_version;

fastcgi_param  SERVER_ADDR        $server_addr;
fastcgi_param  SERVER_PORT        $server_port;
fastcgi_param  SERVER_NAME        $server_name;

# PHP only, required if PHP was built with --enable-force-cgi-redirect
fastcgi_param  REDIRECT_STATUS    200;

fastcgi_param  HTTP_PROXY         "";
"""

SCGI_PARAMS = """
scgi_param  REQUEST_METHOD     $request_method;
scgi_param  REQUEST_URI        $request_uri;
scgi_param  QUERY_STRING       $query_string;
scgi_param  CONTENT_TYPE       $content_type;

scgi_param  DOCUMENT_URI       $document_uri;
scgi_param  DOCUMENT_ROOT      $document_root;
scgi_param  SCGI               1;
scgi_param  SERVER_PROTOCOL    $server_protocol;
scgi_param  REQUEST_SCHEME     $scheme;
scgi_param  HTTPS              $https if_not_empty;

scgi_param  SERVER_PORT        $server_port;
scgi_param  SERVER_NAME        $server_name;

scgi_param  HTTP_PROXY         "";
"""

UWSGI_PARAMS = """
uwsgi_param  QUERY_STRING       $query_string;
uwsgi_param  REQUEST_METHOD     $request_method;
uwsgi_param  CONTENT_TYPE       $content_type;
uwsgi_param  CONTENT_LENGTH     $content_length;

uwsgi_param  REQUEST_URI        $request_uri;
uwsgi_param  PATH_INFO          $document_uri;
uwsgi_param  DOCUMENT_ROOT      $document_root;
uwsgi_param  SERVER_PROTOCOL    $server_protocol;
uwsgi_param  REQUEST_SCHEME     $scheme;
uwsgi_param  HTTPS              $https if_not_empty;

uwsgi_param  SERVER_PORT        $server_port;
uwsgi_param  SERVER_NAME        $server_name;

uwsgi_param  HTTP_PROXY         "";
"""


class ConfigBuilder(LogMixIn, OnDemandMixIn):
    def __init__(self, config, svc):
        self._config = config
        self._nginx_conf = self._ext.copy.deepcopy(
            svc['tune'].get('config', {}))
        self._svc = svc
        self._remote_addr_var = '$remote_addr'
        self._remote_port_var = '$remote_port'
        self._x_forwarded_for_var = '$proxy_add_x_forwarded_for'

        env = config['env']
        nginx_bin = env['nginxBin']

        #---
        if self._detect.isMacOS():
            brew_dir = env.get('brewDir', '')

            if brew_dir:
                self._prefix = brew_dir
            else:
                self._prefix = '/usr/local'
        else:
            self._prefix = ''

        #---
        ver = self._executil.callExternal(
            [nginx_bin, '-v'],
            verbose=False, merge_stderr=True)
        ver = ver.split('/')[1].strip().split('.')
        ver = [int(v) for v in ver]
        self._version = tuple(ver)

    def build(self, name_id, pid_file, tmp_path):
        config = self._config
        deploy = config['deploy']
        svc = self._svc
        svc_tune = svc['tune']
        conf = self._nginx_conf
        cid_tune = svc_tune.get('cid', {})

        # Global
        #---
        #conf['user'] = '{0} {1}'.format(deploy['user'], deploy['group'])
        conf['worker_processes'] = svc_tune['maxCpuCount']
        conf.setdefault('error_log', 'stderr error')
        conf['worker_rlimit_nofile'] = svc_tune['maxFD']
        conf['daemon'] = 'off'
        conf['pcre_jit'] = 'on'
        conf['working_directory'] = config['deployDir']
        conf['timer_resolution'] = '10ms'
        conf['pid'] = pid_file

        # Events
        #---
        events = conf.setdefault('events', OrderedDict())
        events['worker_connections'] = int(
            svc_tune['maxFD'] // svc_tune['maxCpuCount'])

        # HTTP
        #---
        http = conf.setdefault('http', OrderedDict())
        http.setdefault('-types', MIME_TYPES)
        http.setdefault('default_type', 'application/octet-stream')
        http.setdefault('access_log', 'off')
        http.setdefault('log_not_found', 'off')

        #
        temp_path_opt = http.setdefault(
            'client_body_temp_path', '{0} 1 2'.format(tmp_path))
        http.setdefault('proxy_temp_path', temp_path_opt)
        http.setdefault('fastcgi_temp_path', temp_path_opt)
        http.setdefault('uwsgi_temp_path', temp_path_opt)
        http.setdefault('scgi_temp_path', temp_path_opt)
        #
        http.setdefault('proxy_buffering', 'off')
        http.setdefault('proxy_request_buffering', 'off')
        http.setdefault('proxy_max_temp_file_size', '0')
        http.setdefault('proxy_next_upstream',
                        'error timeout http_502 non_idempotent')
        #
        http.setdefault('fastcgi_buffering', 'on')
        http.setdefault('fastcgi_request_buffering', 'off')
        http.setdefault('fastcgi_max_temp_file_size', '128k')
        http.setdefault('fastcgi_next_upstream', 'error')
        #
        if not self._detect.isMacOS():
            http.setdefault('aio', 'threads')
            if self._version >= (1, 9, 13):
                http.setdefault('aio_write', 'on')
        #
        http.setdefault('server_tokens', 'off')

        http['map $http_upgrade $connection_upgrade'] = {
            'default': 'upgrade',
            "''": "''",
        }

        # The only vhost
        #---
        server = http.setdefault('server', OrderedDict())
        server['root'] = self._pathutil.safeJoin(config['wcDir'], svc['path'])
        server.setdefault('gzip_static', 'on')
        server.setdefault('server_name', '_')

        socket_type = svc_tune['socketType']

        if socket_type == 'unix':
            listen = 'unix:{0}'.format(svc_tune['socketPath'])
            server['set_real_ip_from'] = 'unix:'

        elif socket_type == 'tcp':
            listen = '{0}:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])
        elif socket_type == 'tcp6':
            listen = '[{0}]:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])
        else:
            self._errorExit(
                'Unsupported socket type "{0}" for "{1}"'.format(socket_type, name_id))

        listen += ' '
        listenOptions = 'default_server'

        if self._detect.isLinux() and socket_type != 'unix':
            listenOptions += ' deferred'

        listen += cid_tune.get('listenOptions', listenOptions)

        if cid_tune.get('proxyProtocol', False):
            listen += ' proxy_protocol'
            server['real_ip_header'] = 'proxy_protocol'

            if socket_type != 'unix':
                trusted = config['env'].get('nginxTrustedProxy', '0.0.0.0/0')
                server['set_real_ip_from'] = trusted.split()

        server['listen'] = listen

        # Location per each webapp
        #---
        webcfg = config.get('webcfg', {})
        mounts = webcfg.get('mounts', {}).copy()
        autoServices = deploy['autoServices']
        main = webcfg.get('main', None)
        webroot = webcfg.get('root', None)

        if webroot != svc['path']:
            self._errorExit('.webcfg.root mismatches nginx entrypoint .path')

        if main:
            mounts.setdefault('/', {})
            mounts['/']['app'] = main

        def process_static_tune(loc, tune):
            loc['index'] = tune.get('index', 'index.html')
            loc['autoindex'] = 'on' if tune.get('autoindex', False) else 'off'
            loc['etag'] = 'on' if tune.get('etag', False) else 'off'
            loc['expires'] = tune.get('expires', 'max')

            loc['gzip'] = 'on' if tune.get('gzip', False) else 'off'
            loc['gzip_static'] = 'on' if tune.get(
                'staticGzip', False) else 'off'

        for (prefix, info) in mounts.items():

            if not isinstance(info, dict):
                info = {'app': info}

            app = info.get('app', None)
            path_tune = info.get('tune', {})

            if app:
                try:
                    instances = autoServices[app]
                    appsvc = config['entryPoints'][app]
                except KeyError:
                    self._errorExit(
                        'Missing "autoServices" for "{0}" entry point'.format(app))

                if appsvc.get('tool', None) == 'nginx':
                    continue

                try:
                    protocol = instances[0]['socketProtocol']
                except KeyError:
                    self._errorExit(
                        'Missing "socketProtocol" for "{0}" entry point'.format(app))

                if protocol == 'http':
                    upstream, location = self._proxyHttp(
                        app, appsvc, instances)
                elif protocol == 'fcgi':
                    upstream, location = self._proxyFcgi(
                        app, appsvc, instances)
                elif protocol == 'scgi':
                    upstream, location = self._proxyScgi(
                        app, appsvc, instances)
                elif protocol == 'uwsgi':
                    upstream, location = self._proxyUwsgi(
                        app, appsvc, instances)
                else:
                    self._errorExit(
                        'Not supported protocol "{0}" for "{1}" entry point'.format(protocol, app))

                http['upstream {0}'.format(app)] = upstream

                serve_static = info.get('static', False)

                if serve_static:
                    server['location @main'] = location
                    path_location = {
                        'try_files': '$uri @main',
                        'disable_symlinks': 'if_not_owner',
                    }
                    process_static_tune(path_location, path_tune)
                else:
                    path_location = location
            else:
                path_location = {
                    'disable_symlinks': 'if_not_owner',
                }
                process_static_tune(path_location, path_tune)

            server['location {0}'.format(prefix)] = path_location

        if '/' not in mounts:
            if webroot:
                server['location /'] = {
                    'disable_symlinks': 'if_not_owner',
                }
            else:
                server['location /'] = {
                    'deny': 'all',
                }

        return self.getTextConfig()

    def _upstreamCommon(self, app, svc, instances, keepalive=False):
        svc_tune = svc['tune']
        cid_tune = svc_tune.get('cid', {})
        zone_size = cid_tune.get('upstreamZoneSize', '64k')
        fail_timeout = cid_tune.get('upstreamFailTimeout', '0')
        keep_alive_percent = cid_tune.get('upstreamKAPercent', 25)
        queue = cid_tune.get('upstreamQueue', None)

        upstream = OrderedDict()

        if queue:
            upstream['queue'] = queue
            upstream['zone'] = 'upstreams {0}'.format(zone_size)
            upstream['hash'] = '$binary_remote_addr consistent'
        else:
            upstream['least_conn'] = ''

        if keepalive and keep_alive_percent > 0:
            ka_conn = 0
            for v in instances:
                ka_conn += v['maxConnections']
            ka_conn = int(ka_conn * keep_alive_percent // 100)

            if ka_conn:
                upstream['keepalive'] = ka_conn

        for v in instances:
            options = []
            if queue:
                options.append('max_conns={0}'.format(v['maxConnections']))

            options.append('max_fails=0')
            options.append('fail_timeout={0}'.format(fail_timeout))

            socket_type = v['socketType']

            if socket_type == 'unix':
                socket = 'unix:{0}'.format(v['socketPath'])
            elif socket_type == 'tcp':
                sock_addr = v['socketAddress']

                if sock_addr == '0.0.0.0':
                    sock_addr = '127.0.0.1'

                socket = '{0}:{1}'.format(sock_addr, v['socketPort'])
            elif socket_type == 'tcp6':
                sock_addr = v['socketAddress']

                if sock_addr == '::':
                    sock_addr = '::1'

                socket = '[{0}]:{1}'.format(sock_addr, v['socketPort'])
            else:
                self._errorExit(
                    'Unsupported socket type "{0}" for "{1}"'.format(socket_type, app))

            upstream['server {0}'.format(socket)] = ' '.join(options)

        return upstream

    def _svcBodyLimit(self, instances):
        return self._configutil.parseMemory(instances[0]['maxRequestSize'])

    def _proxyHttp(self, app, svc, instances):
        upstream = self._upstreamCommon(app, svc, instances, True)
        location = OrderedDict()
        location['proxy_pass'] = 'http://{0}'.format(app)
        location['proxy_http_version'] = '1.1'
        location['proxy_set_header Upgrade'] = '$http_upgrade'
        location['proxy_set_header Connection'] = '$connection_upgrade'
        location['proxy_set_header Host'] = '$host'
        location['proxy_set_header X-Real-IP'] = self._remote_addr_var
        location['proxy_set_header X-Forwarded-For'] = self._x_forwarded_for_var
        location['proxy_set_header X-Forwarded-Proto'] = '$scheme'
        location['proxy_set_header Proxy'] = '""'
        location['proxy_cache_bypass'] = '$http_upgrade'
        location['proxy_next_upstream_tries'] = len(instances)
        location['client_max_body_size'] = self._svcBodyLimit(instances)

        return upstream, location

    def _proxyFcgi(self, app, svc, instances):
        upstream = self._upstreamCommon(app, svc, instances, True)
        location = OrderedDict()
        location['fastcgi_pass'] = app
        location['fastcgi_keep_conn'] = 'on'
        location['fastcgi_ignore_client_abort'] = 'on'
        location['fastcgi_intercept_errors'] = 'on'
        location['fastcgi_next_upstream_tries'] = len(instances)
        location['fastcgi_read_timeout'] = '600s'
        location['client_max_body_size'] = self._svcBodyLimit(instances)

        location['-fastcgi-params'] = FASTCGI_PARAMS
        location['fastcgi_param REMOTE_ADDR'] = self._remote_addr_var
        location['fastcgi_param REMOTE_PORT'] = self._remote_port_var
        location['fastcgi_param SCRIPT_FILENAME'] = self._pathutil.safeJoin(
            self._config['deployDir'], 'current', svc['path'])

        return upstream, location

    def _proxyScgi(self, app, svc, instances):
        upstream = self._upstreamCommon(app, svc, instances)
        location = OrderedDict()
        location['scgi_pass'] = app
        location['client_max_body_size'] = self._svcBodyLimit(instances)

        location['-fastcgi-params'] = SCGI_PARAMS
        location['scgi_param REMOTE_ADDR'] = self._remote_addr_var
        location['scgi_param REMOTE_PORT'] = self._remote_port_var

        return upstream, location

    def _proxyUwsgi(self, app, svc, instances):
        upstream = self._upstreamCommon(app, svc, instances)
        location = OrderedDict()
        location['uwsgi_pass'] = app
        location['client_max_body_size'] = self._svcBodyLimit(instances)

        location['-fastcgi-params'] = UWSGI_PARAMS
        location['uwsgi_param REMOTE_ADDR'] = self._remote_addr_var
        location['uwsgi_param REMOTE_PORT'] = self._remote_port_var

        return upstream, location

    def getTextConfig(self):
        def prep_config(section, prefix):
            content = []

            options = filter(lambda x: not isinstance(
                x[1], dict), section.items())
            sections = filter(lambda x: isinstance(
                x[1], dict), section.items())

            for (k, v) in options:
                if k[0] == '-':
                    v = v.split("\n")
                    v = ['{0}{1}'.format(prefix, x) for x in v]
                    v = "\n".join(v)
                    content.append(v)
                elif isinstance(v, list):
                    for lv in v:
                        content.append('{0}{1} {2};'.format(prefix, k, lv))
                else:
                    content.append('{0}{1} {2};'.format(prefix, k, v))

            for (k, v) in sections:
                content.append('')
                v = prep_config(v, prefix + "\t")
                content.append("{0}{1} {{\n{2}\n{0}}}".format(prefix, k, v))

            return "\n".join(content)

        return prep_config(self._nginx_conf, "")



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class futoinTool(RunEnvTool):
    """futoin.json updater as defined in FTN16.

Home: https://github.com/futoin/specs/blob/master/draft/ftn16_cid_tool.md

futoin.json is the only file read by FutoIn CID itself.
"""
    __slots__ = ()
    _FUTOIN_JSON = 'futoin.json'

    def autoDetectFiles(self):
        return self._FUTOIN_JSON

    def initEnv(self, env):
        self._have_tool = True

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self._FUTOIN_JSON, updater)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn


class gradleTool(SdkmanToolMixIn, BuildTool):
    """Gradle Build Tool.

Home: https://gradle.org/

Build targets:
    prepare -> clean
    build -> <default> without explicit target
    package -> dists
    packageGlob -> '*.jar'
Override targets with .config.toolTune.

Requires Java >= 7.
"""
    __slots__ = ()

    def _minJava(self):
        return '7'

    def autoDetectFiles(self):
        return 'build.gradle'

    def envDeps(self, env):
        super(gradleTool, self).envDeps(env)

        gradlew_prop = 'gradle/wrapper/gradle-wrapper.properties'

        if self._ospath.exists(gradlew_prop):
            with open(gradlew_prop, 'r') as f:
                props = f.read()
                env['gradleVer'] = self._ext.re.search(
                    'gradle-([0-9.]+)-bin.zip', props).group(1)

    def onPrepare(self, config):
        target = self._getTune(config, 'prepare', 'clean')
        cmd = [config['env']['gradleBin'],
               '-q', '--no-daemon', target]
        self._executil.callMeaningful(cmd)

    def onBuild(self, config):
        target = self._getTune(config, 'build')

        if target:
            args = [target]
        else:
            args = []

        cmd = [config['env']['gradleBin'],
               '-q', '--no-daemon'] + args
        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'dists')
        cmd = [config['env']['gradleBin'],
               '-q', '--no-daemon', target]
        self._executil.callMeaningful(cmd)

        packageGlob = self._getTune(config, 'packageGlob', 'build/libs/*.jar')
        self._pathutil.addPackageFiles(config, packageGlob)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class gemTool(BuildTool):
    """RubyGems: Find, install, and publish RubyGems.

Home: https://rubygems.org/

If rubyVer is equal to system then gems are installed in
user's folder gemDir.

gemDir is equal to ~/.gem by default.

gemInstallArgs is forcibly set by tool depending on its version.
"""
    __slots__ = ()

    def getDeps(self):
        return ['ruby']

    def uninstallTool(self, env):
        pass

    def envNames(self):
        return ['gemBin', 'gemDir', 'gemInstallArgs']

    def initEnv(self, env):
        ospath = self._ospath
        environ = self._environ
        installArgs = []

        if env['rubyVer'] == self.SYSTEM_VER or env['rubyBinOnly']:
            gemDir = ospath.join(self._pathutil.deployHome(),
                                 '.gem', env['rubyVer'])
            gemDir = env.setdefault('gemDir', gemDir)
            environ['GEM_HOME'] = gemDir
            environ['GEM_PATH'] = gemDir
            environ['GEM_SPEC_CACHE'] = ospath.join(gemDir, 'specs')

            #self._pathutil.addEnvPath('GEM_PATH', gemDir)
            self._pathutil.addBinPath(ospath.join(gemDir, 'bin'), True)
            installArgs += ['--no-user-install', '--no-format-executable']

        super(gemTool, self).initEnv(env)

        if self._have_tool:
            version = self._executil.callExternal(
                [env['gemBin'], '--version'], verbose=False).strip()

            if version >= '2':
                installArgs += ['--no-document']
            else:
                installArgs += ['--no-ri', '--no-rdoc']

            env['gemInstallArgs'] = ' '.join(installArgs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from ..details.nginx import *

__all__ = 'nginxTool'


class nginxTool(RuntimeTool):
    """nginx [engine x], originally written by Igor Sysoev.
Home: http://nginx.org

nginx support is targeted only cases when server is run behind
reverse proxy or level 7 load balancer. It's NOT configured
for frontend cases !!!

Note: nginxVer supports only "stable" or "mainline" trying to setup from nginx.org
repos as far as possible. Please remove OS-supplied nginx, if installed.

Tune possibilities through .tune.config - representing dict-based tree structure:
* global directives
* .events
    * events directives
* .http
    * http-level directives
    * .server - the only vhost
        * server-level directives

It's possible to override automatic variables in some cases.

Additional logic tuning is available through .tune.cid:
* .listenOptions = 'default_server deferred'
* .upstreamZoneSize = 64k - shared memory for upstream zones
* .upstreamFailTimeout = 10 - upstream failure timeout
* .upstreamQueue = None - paramaters to "queue", if set (commercial)
* .upstreamKAPercent = 25 - Keep-Alive percent
* .proxyProtocol = False - enable binary proxy-protocol



Additional notes:
* proxy buffering is disabled and assumed to be part of front- or middle- balancers

"""
    __slots__ = ()

    _GPG_KEY = GPG_KEY

    def envNames(self):
        return ['nginxBin', 'nginxVer', 'nginxBaseUrl', 'nginxTrustedProxy']

    def tuneDefaults(self, env):
        return {
            'minMemory': '4M',
            'connMemory': '32K',
            'connFD': 8,
            'socketTypes': ['unix', 'tcp', 'tcp6'],
            'socketType': 'tcp' if env['type'] == 'dev' else 'unix',
            'socketProtocol': 'http',
            'scalable': False,
            'reloadable': True,
            'multiCore': True,
            'maxRequestSize': '1M',
            'memoryWeight': 10,
        }

    def _installTool(self, env):
        base_url = env['nginxBaseUrl']

        if env['nginxVer'] == 'mainline':
            base_url += '/mainline'

        detect = self._detect

        if detect.isDebian():
            self._install.aptRepo(
                'nginx',
                'deb {0}/debian/ $codename$ nginx'.format(base_url),
                self._GPG_KEY,
                codename_map={
                    'sid': 'stretch',
                    'testing': 'stretch',
                })

        elif detect.isUbuntu():
            self._install.aptRepo(
                'nginx',
                'deb {0}/ubuntu/ $codename$ nginx'.format(base_url),
                self._GPG_KEY,
                codename_map={})

        elif detect.isCentOS():
            dist_ver = self._detect.linuxDistMajorVer()

            self._install.yumRepo(
                'nginx',
                '{0}/centos/{1}/$basearch/'.format(base_url, dist_ver),
                self._GPG_KEY,
                repo_url=True)

        elif detect.isFedora():
            self._install.yumRepo(
                'nginx',
                '{0}/centos/7/$basearch/'.format(base_url),
                self._GPG_KEY,
                repo_url=True)

        elif detect.isRHEL() or detect.isOracleLinux():
            dist_ver = self._detect.linuxDistMajorVer()

            self._install.yumRepo(
                'nginx',
                '{0}/rhel/{1}/$basearch/'.format(base_url, dist_ver),
                self._GPG_KEY,
                repo_url=True)

        elif detect.isSLES():
            sles_ver = self._detect.linuxDistMajorVer()
            self._install.zypperRepo(
                'nginx',
                '{0}/sles/{1}/'.format(base_url, sles_ver),
                self._GPG_KEY,
                yum=True)

        elif detect.isArchLinux():
            if env['nginxVer'] == 'mainline':
                self._install.pacman('nginx-mainline')
            else:
                self._install.pacman('nginx')
            return

        self._install.debrpm('nginx')
        self._install.apk('nginx')
        self._install.brew('nginx')

    def initEnv(self, env, bin_name=None):
        env.setdefault('nginxBaseUrl', 'http://nginx.org/packages')
        env.setdefault('nginxVer', 'stable')

        super(nginxTool, self).initEnv(env, bin_name)

        sbin_nginx = '/usr/sbin/nginx'

        if not self._have_tool and self._ospath.exists(sbin_nginx):
            env['nginxBin'] = sbin_nginx
            self._have_tool = True

    def onPreConfigure(self, config, runtime_dir, svc, cfg_svc_tune):
        ospath = self._ospath
        deploy = config['deploy']

        name_id = '{0}-{1}'.format(svc['name'], svc['instanceId'])

        conf_file = 'nginx-{0}.conf'.format(name_id)
        conf_file = ospath.join(runtime_dir, conf_file)

        pid_file = 'nginx-{0}.pid'.format(name_id)
        pid_file = ospath.join(runtime_dir, pid_file)

        cfg_svc_tune['nginxConf'] = conf_file

        # max of all proxied connections
        max_request_size = self._configutil.parseMemory(
            svc['tune']['maxRequestSize'])
        for sl in deploy['autoServices'].values():
            for s in sl:
                max_request_size = max(
                    max_request_size,
                    self._configutil.parseMemory(s['maxRequestSize'])
                )
        cfg_svc_tune['maxRequestSize'] = self._configutil.toMemory(
            max_request_size)
        #
        tmp_path = ospath.join(deploy['tmpDir'], 'nginx')
        self._pathutil.mkDir(tmp_path)

        #
        conf_builder = ConfigBuilder(config, svc)
        nginx_conf = conf_builder.build(name_id, pid_file, tmp_path)
        self._pathutil.writeTextFile(conf_file, nginx_conf)

        # Verify
        #---
        env = config['env']
        self._executil.callExternal([
            env['nginxBin'],
            '-q', '-t',
            '-p', runtime_dir,
            '-c', conf_file,
        ])

    def onRun(self, config, svc, args):
        env = config['env']
        deploy = config['deploy']
        self._executil.callInteractive([
            env['nginxBin'],
            '-p', deploy['runtimeDir'],
            '-c', svc['tune']['nginxConf'],
        ] + args)

    def onStop(self, config, pid, tune):
        env = config['env']
        deploy = config['deploy']
        self._executil.callExternal([
            env['nginxBin'],
            '-s', 'stop',
            '-p', deploy['runtimeDir'],
            '-c', tune['nginxConf'],
        ])

    def onReload(self, config, pid, tune):
        env = config['env']
        deploy = config['deploy']
        self._executil.callExternal([
            env['nginxBin'],
            '-s', 'reload',
            '-p', deploy['runtimeDir'],
            '-c', tune['nginxConf'],
        ])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .gemtoolmixin import GemToolMixIn


class bundlerTool(GemToolMixIn, BuildTool):
    """Bundler: The best way to manage a Ruby application's gems.

Home: http://bundler.io/

Note:
1. It will use per-project BUNDLE_PATH=vendor/bundle by default
2. In deployment with missing vendor/bundle, it uses per-deployment BUNDLE_PATH
3. If run outside of project/deployment, standard ~/.bundle is used
"""
    __slots__ = ()

    def envNames(self):
        return ['bundlePath', 'bundlerVer']

    def autoDetectFiles(self):
        return 'Gemfile'

    def initEnv(self, env):
        ospath = self._ospath

        if ospath.exists('vendor/bundle'):
            # packed deps
            bundlePath = self._ospath.realpath('vendor/bundle')
        elif ospath.exists('current/vendor/bundle'):
            # packed deps
            bundlePath = self._ospath.realpath('current/vendor/bundle')
        elif not ospath.exists('Gemfile'):
            # global / deployment
            bundlePath = ospath.join(self._pathutil.deployHome(), '.bundle')
        else:
            # per project
            bundlePath = self._ospath.realpath('vendor/bundle')

        bundlePath = env.setdefault('bundlePath', bundlePath)
        self._environ['BUNDLE_PATH'] = bundlePath

        super(bundlerTool, self).initEnv(env, 'bundle')

    def onPrepare(self, config):
        env = config['env']

        # Dirty hack
        #---
        bundlerTools = env.get('bundlerTools', {})
        do_bundler_hack = len(bundlerTools) > 0

        for (k, v) in bundlerTools.items():
            tcmd = [env['bundlerBin'], 'add', k]

            if v:
                tcmd.append('--version={0}'.format(v))

            try:
                self._executil.callExternal(tcmd, suppress_fail=True)
            except Exception as e:
                self._warn(str(e))

        if len(bundlerTools) > 0:
            cmd = [env['bundlerBin'], 'install']
            self._executil.callExternal(cmd)

        # Main install
        #---
        cmd = [env['bundlerBin'], 'install']

        if self._ospath.exists('Gemfile.lock'):
            cmd.append('--deployment')

        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        cmd = [config['env']['bundlerBin'], 'install',
               '--deployment', '--clean']
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn
from .curltoolmixin import CurlToolMixIn


class rvmTool(BashToolMixIn, CurlToolMixIn, RunEnvTool):
    """Ruby Version Manager.

Home: https://rvm.io/

Stable RVM is used by default.
"""
    __slots__ = ()

    RVM_LATEST = 'stable'
    RVM_GPG_KEYS = [
        '409B6B1796C275462A1703113804BB82D39DC0E3',
        '7D2BAF1CF37B13E2069D6956105BD0E739499BDB',
    ]
    RVM_GET = 'https://get.rvm.io'

    def getDeps(self):
        return (
            ['gpg', 'tar', 'gzip'] +
            BashToolMixIn.getDeps(self) +
            CurlToolMixIn.getDeps(self))

    def ensureGpgKeys(self, env):
        rvm_gpg_keys = env.get('rvmGpgKeys', '')

        if rvm_gpg_keys != '':
            rvm_gpg_keys = rvm_gpg_keys.split(' ')
        else:
            rvm_gpg_keys = self.RVM_GPG_KEYS

        self._executil.callExternal([
            env['gpgBin'], '--keyserver', env['gpgKeyServer'],
            '--recv-keys'] + rvm_gpg_keys,
            suppress_fail=True)

    def _installTool(self, env):
        self.ensureGpgKeys(env)

        rvm_dir = env['rvmDir']
        rvm_get = env.get('rvmGet', self.RVM_GET)

        environ = self._environ
        environ['rvm_user_install_flag'] = '1'
        environ['rvm_auto_dotfiles_flag'] = '0'

        installer = self._callCurl(env, [rvm_get])

        # AlpineLinux - required for uninstall
        self._install.apk('procps')

        self._callBash(
            env,
            bash_args=['--', env['rvmVer'], '--path', rvm_dir],
            input=installer)

    def _updateTool(self, env):
        self._executil.callExternal([env['rvmBin'], 'get', env['rvmVer']])

    def uninstallTool(self, env):
        try:
            self._executil.callExternal([env['rvmBin'], 'implode', '--force'])
        except:
            pass

        rvm_dir = env['rvmDir']
        self._pathutil.rmTree(rvm_dir)
        self._have_tool = False

    def envNames(self):
        return ['rvmVer', 'rvmDir', 'rvmBin', 'rvmGet', 'rvmGpgKey']

    def initEnv(self, env):
        ospath = self._ospath
        environ = self._environ

        for v in ['rvm_path', 'rvm_bin_path', 'rvm_prefix', 'rvm_version']:
            try:
                del environ[v]
            except:
                pass

        env.setdefault('rvmVer', self.RVM_LATEST)
        rvm_dir = ospath.join(environ['HOME'], '.rvm')
        rvm_dir = env.setdefault('rvmDir', rvm_dir)
        rvm_bin_dir = ospath.join(rvm_dir, 'bin')
        rvm_bin = env['rvmBin'] = ospath.join(rvm_bin_dir, 'rvm')
        env['rvmInit'] = ospath.join(rvm_dir, 'scripts', 'rvm')
        self._have_tool = ospath.exists(rvm_bin)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .sdkmantoolmixin import SdkmanToolMixIn


class antTool(SdkmanToolMixIn, BuildTool):
    """Ant build tool for Java applications.

Home: http://ant.apache.org/

The tool assumes the following targets: clean, compile, jar, run

Ant is setup through SDKMan!

Note: If detected Java version is less than 8 then Ant 1.9.8 is used.

Build targets:
    prepare -> clean
    build -> compile
    package -> jar
Override targets with .config.toolTune.

"""
    __slots__ = ()

    def autoDetectFiles(self):
        return ['build.xml']

    def initEnv(self, env):
        if self._javaVersion(env) < 8:
            env['antVer'] = '1.9.8'

        super(antTool, self).initEnv(env)

    def _callAnt(self, config, target):
        cmd = [config['env']['antBin'], target]
        self._executil.callMeaningful(cmd)

    def onPrepare(self, config):
        target = self._getTune(config, 'prepare', 'clean')
        self._callAnt(config, target)

    def onBuild(self, config):
        target = self._getTune(config, 'build', 'compile')
        self._callAnt(config, target)

    def onPackage(self, config):
        target = self._getTune(config, 'package', 'jar')
        self._callAnt(config, target)

        self._pathutil.addPackageFiles(config, 'build/jar/*.jar')

    def onRunDev(self, config):
        self._callAnt(config, 'run')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class bowerTool(NpmToolMixIn, BuildTool):
    """Bower: a package manager for the web.

Home: https://bower.io/
"""
    __slots__ = ()

    BOWER_JSON = 'bower.json'

    def autoDetectFiles(self):
        return self.BOWER_JSON

    def loadConfig(self, config):
        content = self._pathutil.loadJSONConfig(self.BOWER_JSON)
        if content is None:
            return

        f = 'name'
        if f in content and f not in config:
            config[f] = content[f]

    def updateProjectConfig(self, config, updates):
        def updater(json):
            f = 'name'
            if f in updates:
                json[f] = updates[f]

            # version is deprecated
            if 'version' in json:
                del json['version']

        return self._pathutil.updateJSONConfig(self.BOWER_JSON, updater)

    def onPrepare(self, config):
        bowerBin = config['env']['bowerBin']
        self._executil.callExternal([bowerBin, 'install'])

    def onPackage(self, config):
        # Bower does not remove dev deps by itself
        self._pathutil.rmTree('bower_components')

        bowerBin = config['env']['bowerBin']
        self._executil.callExternal([bowerBin, 'install', '--production'])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext


def emerge(packages):
    if not _ext.detect.isEmerge():
        return

    emerge = _ext.pathutil.which('emerge')

    packages = _ext.configutil.listify(packages)

    _ext.executil.trySudoCall(
        [emerge] + packages,
        errmsg='you may need to install the build deps manually !'
    )


def emergeDepsOnly(packages):
    packages = _ext.configutil.listify(packages)
    emerge(['--onlydeps'] + packages)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext
from .. import log as _log


def deb(packages):
    if not _ext.detect.isDeb():
        return

    apt_get = _ext.pathutil.which('apt-get')

    packages = _ext.configutil.listify(packages)

    _ext.os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
    _ext.executil.trySudoCall(
        [apt_get, 'install', '-y',
            '--no-install-recommends',
            '-o', 'Dpkg::Options::=--force-confdef',
            '-o', 'Dpkg::Options::=--force-confold'] + packages,
        errmsg='you may need to install the packages manually !'
    )


def aptRepo(name, entry, gpg_key=None, codename_map=None, repo_base=None):
    if not _ext.detect.isDeb():
        return

    deb([
        'software-properties-common',
        'apt-transport-https',
        'ca-certificates',
        'lsb-release',
    ])
    apt_add_repository = _ext.pathutil.which('apt-add-repository')

    if not apt_add_repository:
        return

    if gpg_key:
        tmp_dir = _ext.pathutil.tmpCacheDir(prefix='cidgpg')
        tf = _ext.ospath.join(tmp_dir, 'key.gpg')
        _ext.pathutil.writeBinaryFile(tf, gpg_key)

        _ext.executil.trySudoCall(
            ['apt-key', 'add', tf],
            errmsg='you may need to import the PGP key manually!'
        )

        _ext.os.remove(tf)

    codename = _ext.detect.osCodeName()

    if codename_map:
        try:
            repo_info = _ext.urllib.urlopen(
                '{0}/{1}'.format(repo_base, codename)).read()
        except:
            fallback_codename = codename_map.get(codename, codename)
            _log.warn('Fallback to codename: {0}'.format(
                fallback_codename))
            codename = fallback_codename

    entry = entry.replace('$codename$', codename)

    _ext.executil.trySudoCall(
        [apt_add_repository, '--yes', entry],
        errmsg='you may need to add the repo manually!'
    )

    _ext.executil.trySudoCall(
        ['apt-get', 'update'],
        errmsg='you may need to update APT cache manually!'
    )

def dpkg(env, name, url):
    if not _ext.detect.isDeb():
        return

    pathutil = _ext.pathutil
    apt_cache = pathutil.which('apt-cache')

    if apt_cache and _ext.executil.callExternal([apt_cache, 'show', name], suppress_fail=True):
        _ext.install.deb(name)
    else:
        dpkg = pathutil.which('dpkg')

        if _ext.executil.callExternal([dpkg, '-s', name], suppress_fail=True):
            return

        cache_file = pathutil.cacheDownloadFile(env, url)
        _ext.executil.trySudoCall([dpkg, '-i', cache_file])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .data import DataSlots


class DeployMixIn(DataSlots):
    __slots__ = ()

    __VCS_CACHE_DIR = 'vcs'

    def __init__(self):
        super(DeployMixIn, self).__init__()
        self._current_dir = None
        self._devserve_mode = False

    def _redeployExit(self, deploy_type):
        self._warn(deploy_type + " has been already deployed. Use --redeploy.")
        self._sys.exit(0)

    def _rms_deploy(self, rms_pool, package=None):
        self._requireDeployLock()

        ospath = self._ospath
        os = self._os
        config = self._config
        rmstool = self._getRmsTool()

        # Find out package to deploy
        self._info('Finding package in RMS')
        package_list = rmstool.rmsGetList(config, rms_pool, package)

        if package:
            package_list = self._ext.fnmatch.filter(package_list, package)

        if not package_list:
            self._errorExit("No package found")

        package = self._versionutil.latest(package_list)
        self._info('Found package {0}'.format(package))

        # cleanup first, in case of incomplete actions
        self._info('Pre-cleanup of deploy directory')
        self._deployCleanup([package])

        # Prepare package name components
        package_basename = ospath.basename(package)
        (package_noext, package_ext) = ospath.splitext(package_basename)

        # Check if already deployed:
        if ospath.exists(package_noext):
            if config['reDeploy']:
                self._warn('Forcing re-deploy of the package')
            else:
                self._redeployExit('Package')

        # Retrieve package, if not available
        if not ospath.exists(package_basename):
            self._info('Retrieving the package')
            package_list = [package]
            package_list = rmstool.rmsProcessChecksums(
                config, rms_pool, package_list)
            rmstool.rmsRetrieve(config, rms_pool, package_list)

        package_noext_tmp = package_noext + '.tmp'

        # Prepare temporary folder
        os.mkdir(package_noext_tmp)

        # Unpack package to temporary folder
        self._info('Extracting the package')
        env = config['env']

        tool = None
        tool_args = None
        tar_common_args = ['--strip-components=1', '-C', package_noext_tmp]

        if package_ext == '.txz':
            pkg_tool = self._getTarTool('xz')
            tool_args = ['xJf', package_basename] + tar_common_args
        elif package_ext == '.tbz2':
            pkg_tool = self._getTarTool('bzip2')
            tool_args = ['xjf', package_basename] + tar_common_args
        elif package_ext == '.tgz':
            pkg_tool = self._getTarTool('gzip')
            tool_args = ['xzf', package_basename] + tar_common_args
        elif package_ext == '.tar':
            pkg_tool = self._getTarTool()
            tool_args = ['xf', package_basename] + tar_common_args
        else:
            self._errorExit('Not supported package format: ' + package_ext)

        pkg_tool.onExec(env, tool_args, False)

        # Common processing
        self._deployCommon(package_noext_tmp, package_noext, [package])

    def _vcsref_deploy(self, vcs_ref):
        self._requireDeployLock()

        ospath = self._ospath
        os = self._os
        config = self._config
        vcstool = self._getVcsTool()

        # Find out package to deploy
        self._info('Getting the latest revision of {0}'.format(vcs_ref))
        vcs_cache = ospath.realpath(self.__VCS_CACHE_DIR)
        rev = vcstool.vcsGetRefRevision(config, vcs_cache, vcs_ref)

        if not rev:
            self._errorExit("No VCS refs found")

        target_dir = vcs_ref.replace(os.sep, '_').replace(':', '_')
        target_dir += '__' + rev

        # cleanup first, in case of incomplete actions
        self._info('Pre-cleanup of deploy directory')
        self._deployCleanup([self.__VCS_CACHE_DIR, target_dir])

        # Check if already deployed:
        if ospath.exists(target_dir):
            if config['reDeploy']:
                self._warn('Forcing re-deploy of the VCS ref')
            else:
                self._redeployExit('VCS ref')

        # Retrieve tag
        self._info('Retrieving the VCS ref')
        target_tmp = target_dir + '.tmp'
        # Note: acceptable race condition is possible: vcs_ref
        # may get updated after we get its revision and before
        # we do actual export
        vcstool.vcsExport(config, vcs_cache, vcs_ref, target_tmp)

        # Common processing
        self._deployCommon(target_tmp, target_dir, [self.__VCS_CACHE_DIR])

    def _vcstag_deploy(self, vcs_ref):
        self._requireDeployLock()

        ospath = self._ospath
        os = self._os
        config = self._config
        vcstool = self._getVcsTool()

        # Find out package to deploy
        self._info('Finding tag in VCS')
        vcs_cache = ospath.realpath(self.__VCS_CACHE_DIR)
        tag_list = vcstool.vcsListTags(config, vcs_cache, vcs_ref)

        if vcs_ref:
            tag_list = self._ext.fnmatch.filter(tag_list, vcs_ref)

        if not tag_list:
            self._errorExit("No tags found")

        vcs_ref = self._versionutil.latest(tag_list)
        target_dir = vcs_ref.replace(os.sep, '_').replace(':', '_')
        self._info('Found tag {0}'.format(vcs_ref))

        # cleanup first, in case of incomplete actions
        self._info('Pre-cleanup of deploy directory')
        self._deployCleanup([self.__VCS_CACHE_DIR, target_dir])

        # Check if already deployed:
        if ospath.exists(target_dir):
            if config['reDeploy']:
                self._warn('Forcing re-deploy of the VCS tag')
            else:
                self._redeployExit('VCS tag')

        # Retrieve tag
        self._info('Retrieving the VCS tag')
        vcs_ref_tmp = target_dir + '.tmp'
        vcstool.vcsExport(config, vcs_cache, vcs_ref, vcs_ref_tmp)

        # Common processing
        self._deployCommon(vcs_ref_tmp, target_dir, [self.__VCS_CACHE_DIR])

    def _deploy_setup(self):
        self._deployConfig()

    def _deployCommon(self, tmp, dst, cleanup_whitelist):
        self._requireDeployLock()

        ospath = self._ospath
        os = self._os
        stat = self._ext.stat

        self._current_dir = tmp
        config = self._config
        persistent_dir = ospath.realpath(
            config['env'].get('persistentDir', 'persistent'))

        # Predictable change of CWD
        self._overrides['wcDir'] = config['wcDir'] = ospath.realpath(tmp)
        self._processWcDir()
        config = self._config

        # Update tools
        if not self._detect.isDisabledToolsSetup(config['env']):
            self._info('Updating tools')
            self.tool_update(None)

        # Setup persistent folders
        self._info('Setting up read-write directories')

        if not ospath.exists(persistent_dir):
            os.makedirs(persistent_dir)

        wfile_wperm = stat.S_IRUSR | stat.S_IRGRP | stat.S_IWUSR | stat.S_IWGRP
        wdir_wperm = wfile_wperm | stat.S_IXUSR | stat.S_IXGRP

        for def_dir in config.get('persistent', []):
            persist_dst = ospath.join(persistent_dir, def_dir)
            sym_target = ospath.relpath(persist_dst, ospath.dirname(def_dir))
            self._info('{0} -> {1}'.format(def_dir, persist_dst),
                       'Making persistent: ')

            if ospath.isdir(def_dir):
                self._info('copy tree', ' >> ')
                self._ext.dir_util.copy_tree(
                    def_dir, persist_dst,
                    preserve_symlinks=1, preserve_mode=0
                )
                self._pathutil.rmTree(def_dir)
                os.symlink(sym_target, def_dir)
                self._pathutil.chmodTree(
                    persist_dst, wdir_wperm, wfile_wperm, False)
            elif ospath.isfile(def_dir):
                self._info('moving file', ' >> ')
                self._ext.shutil.move(def_dir, persist_dst)
                os.unlink(def_dir)
                os.symlink(sym_target, def_dir)
                os.chmod(persist_dst, wfile_wperm)
            elif ospath.isdir(persist_dst):
                self._info('symlink existing dir', ' >> ')
                os.symlink(sym_target, def_dir)
                self._pathutil.chmodTree(
                    persist_dst, wdir_wperm, wfile_wperm, False)
            else:
                self._info('create & symlink dir', ' >> ')
                os.makedirs(persist_dst, wdir_wperm)
                os.symlink(sym_target, def_dir)

        # Build
        if config.get('deployBuild', False):
            self._info('Building project in deployment')
            self.prepare(None)
            self.build()

        # Symlink .env file, if exists in deploy folder
        if ospath.exists('../.env'):
            os.symlink('../.env', '.env')

        # Complete migration
        self.migrate()

        # return back
        self._processDeployDir()
        config = self._config

        # Setup read-only permissions
        self._info('Setting up read-only permissions')

        file_perm = stat.S_IRUSR | stat.S_IRGRP
        dir_perm = file_perm | stat.S_IXUSR | stat.S_IXGRP
        self._pathutil.chmodTree(tmp, dir_perm, file_perm, True)

        for wdir in config.get('writable', []):
            wdir = ospath.join(tmp, wdir)
            self._pathutil.chmodTree(wdir, wdir_wperm, wfile_wperm, True)

        # Setup services
        self._deployConfig()

        # Move in place
        self._info('Switching current deployment')
        if ospath.exists(dst):
            # re-deploy case
            os.chmod(dst, stat.S_IRWXU)  # macOS
            os.rename(dst, dst + '.tmprm')

        os.chmod(tmp, stat.S_IRWXU)  # macOS
        os.rename(tmp, dst)
        os.chmod(dst, dir_perm)

        os.symlink(dst, 'current.tmp')
        os.rename('current.tmp', 'current')
        self._current_dir = None

        # Re-run
        self._reloadServices()

        # Cleanup old packages and deploy dirs
        self._info('Post-cleanup of deploy directory')
        self._deployCleanup(cleanup_whitelist)

    def _deployCleanup(self, whitelist):
        self._requireDeployLock()

        ospath = self._ospath
        os = self._os

        if ospath.exists('current'):
            whitelist.append(ospath.basename(os.readlink('current')))

        whitelist += [
            'current',
            'persistent',
            self._FUTOIN_JSON,
        ]

        for f in os.listdir('.'):
            (f_noext, f_ext) = ospath.splitext(f)

            if f[0] == '.' or f in whitelist:
                continue

            self._pathutil.rmTree(f)

    def _deployConfig(self):
        self._requireDeployLock()

        self._rebalanceServices()
        self._configServices()

        self._writeDeployConfig()

    def _writeDeployConfig(self):
        self._requireDeployLock()

        config = self._config
        orig_config = self._deploy_config
        new_config = self._exportConfig(orig_config)

        if config.get('rms', None):
            new_config['rms'] = config['rms']
            new_config['rmsRepo'] = config.get('rmsRepo', '')

        if config.get('vcs', None):
            new_config['vcs'] = config['vcs']
            new_config['vcsRepo'] = config.get('vcsRepo', '')

        new_config['deploy'] = config.get('deploy', {})
        new_config['env'] = orig_config.get('env', {})

        self._info('Writing deployment config in {0}'.format(
            self._os.getcwd()))
        self._pathutil.writeJSONConfig(self._FUTOIN_JSON, new_config)

        merged_config = self._exportConfig(config)
        merged_config['deploy'] = new_config['deploy']
        merged_config['env'] = config['env']

        self._pathutil.writeJSONConfig(self._FUTOIN_MERGED_JSON, merged_config)

    def _reloadServices(self):
        self._requireDeployLock()

        # Only service master mode is supported this way
        # Otherwise, deployment invoker is responsible for service reload.
        self._reloadServiceMaster()

    def _rebalanceServices(self):
        self._info('Re-balancing services')
        from ..details.resourcealgo import ResourceAlgo
        ResourceAlgo().configServices(self._config)

    def _configServices(self):
        self._info('Configuring services')
        self._requireDeployLock()

        config = self._config
        service_list = self._configServiceList(config)

        if not service_list:
            return

        deploy = config['deploy']
        runtimeDir = deploy['runtimeDir']
        tmpDir = deploy['tmpDir']

        pathutil = self._pathutil

        # DO NOT use realpath as it may point to "old current"
        config['wcDir'] = pathutil.safeJoin(config['deployDir'], 'current')

        pathutil.mkDir(runtimeDir)
        pathutil.mkDir(tmpDir)

        auto_services = config['deploy']['autoServices']

        from ..runtimetool import RuntimeTool

        for svc in service_list:
            tool = svc['tool']
            t = self._getTool(tool)

            if isinstance(t, RuntimeTool):
                cfg_svc = auto_services[svc['name']][svc['instanceId']]
                t.onPreConfigure(config, runtimeDir, svc, cfg_svc)
            else:
                self._errorExit(
                    'Tool "{0}" for "{1}" is not of RuntimeTool type'
                    .format(tool, svc['name']))

    def _processDeployDir(self):
        os = self._os
        ospath = self._ospath
        stat = self._ext.stat

        os.umask(0o027)

        deploy_dir = self._overrides['deployDir']

        # make sure wcDir is always set to current
        current = self._getDeployCurrent()
        wc_dir = ospath.join(deploy_dir, current)
        self._overrides['wcDir'] = wc_dir

        if not deploy_dir:
            deploy_dir = ospath.realpath('.')
            self._overrides['deployDir'] = deploy_dir

        self._info('Using {0} as deploy directory'.format(deploy_dir))

        if not ospath.exists(deploy_dir):
            self._info('Creating deploy directory')
            os.makedirs(deploy_dir)
            self._deployLock()
            self._deployUnlock()
        else:
            self._checkDeployLock()

        st_mode = os.stat(deploy_dir).st_mode

        if not st_mode & st_mode:
            os.chmod(deploy_dir, st_mode | st_mode)

        if self._config is not None:
            self._info('Re-initializing config')

        self._initConfig()

        if deploy_dir != os.getcwd():
            self._infoLabel('Changing to: ', deploy_dir)
            os.chdir(deploy_dir)

        # some tools may benefit of it
        if not self._devserve_mode:
            self._environ['CID_DEPLOY_HOME'] = deploy_dir

        deploy = self._config['deploy']

        if 'user' not in deploy:
            deploy['user'] = self._ext.pwd.getpwuid(os.geteuid())[0]

        if 'group' not in deploy:
            deploy['group'] = self._ext.grp.getgrgid(os.getegid())[0]

    def _deploy_set_action(self, name, actions):
        dc = self._deploy_config
        dc_actions = dc.setdefault('actions', {})
        dc_actions[name] = actions

    def _deploy_reset_action(self):
        dc = self._deploy_config
        dc['actions'] = {}

    def _deploy_set_persistent(self, paths):
        dc = self._deploy_config
        persistent = set(dc.get('persistent', []))
        persistent.update(set(paths))
        dc['persistent'] = sorted(list(persistent))

    def _deploy_reset_persistent(self):
        dc = self._deploy_config
        dc['persistent'] = []

    def _deploy_set_writable(self, paths):
        dc = self._deploy_config
        writable = set(dc.get('writable', []))
        writable.update(set(paths))
        dc['writable'] = sorted(list(writable))

    def _deploy_reset_writable(self):
        dc = self._deploy_config
        dc['writable'] = []

    def _deploy_set_entrypoint(self, name, tool, path, tune):
        dc = self._deploy_config
        epoints = dc.setdefault('entryPoints', {})
        ep = epoints.setdefault(name, {})
        ep['tool'] = tool
        ep['path'] = path
        ep['tune'] = self._deploy_set_tune_common(ep.get('tune', {}), tune)

    def _deploy_reset_entrypoint(self):
        dc = self._deploy_config
        dc['entryPoints'] = {}

    def _deploy_set_tune_common(self, target, tune):
        if len(tune) == 1 and len(tune[0]) and tune[0][0] == '{':
            return self._configutil.parseJSON(tune[0])
        else:
            for t in tune:
                t = t.split('=', 1)
                tkey = t[0]

                if len(t) == 2 and t[1]:
                    target[tkey] = t[1]
                elif tkey in target:
                    del target[tkey]
            return target

    def _deploy_set_env(self, var, val):
        dc = self._deploy_config
        env = dc.setdefault('env', {})

        if val is not None:
            env[var] = val
        elif var in env:
            del env[var]

    def _deploy_reset_env(self):
        dc = self._deploy_config
        dc['env'] = {}

    def _deploy_set_webcfg(self, var, val):
        dc = self._deploy_config
        webcfg = dc.setdefault('webcfg', {})

        if var == 'mounts':
            mounts = webcfg.setdefault('mounts', {})
            val = val.split('=', 1)
            var = val[0]

            if len(val) == 2 and val[1]:
                mounts[var] = val[1]
            elif var in mounts:
                del mounts[var]

        elif val is not None:
            webcfg[var] = val

        elif var in webcfg:
            del webcfg[var]

    def _deploy_reset_webcfg(self):
        dc = self._deploy_config
        dc['webcfg'] = {}

    def _deploy_set_webmount(self, path, json):
        dc = self._deploy_config
        webcfg = dc.setdefault('webcfg', {})

        mounts = webcfg.setdefault('mounts', {})

        if json is None:
            del mounts[path]
        else:
            mounts[path] = self._configutil.parseJSON(json)

    def _deploy_reset_webmount(self):
        dc = self._deploy_config
        webcfg = dc.setdefault('webcfg', {})

        webcfg['mounts'] = {}

    def _deploy_set_tools(self, tools):
        dc = self._deploy_config

        # set fresh
        dc_tools = {}
        dc['tools'] = dc_tools

        for t in tools:
            v = (t + '=*').split('=')
            v = list(filter(None, v))
            dc_tools[v[0]] = v[1]

    def _deploy_reset_tools(self):
        dc = self._deploy_config
        dc['tools'] = {}

    def _deploy_set_tooltune(self, tool, tune):
        dc = self._deploy_config
        toolTune = dc.setdefault('toolTune', {})
        toolTune[tool] = self._deploy_set_tune_common(
            toolTune.get(tool, {}), tune)

    def _deploy_reset_tooltune(self):
        dc = self._deploy_config
        dc['toolTune'] = {}

    def _getDeployCurrent(self):
        return self._current_dir or 'current'



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class gccTool(RunEnvTool):
    """GNU Compiler Collection.

Home: https://gcc.gnu.org/

If gccBin is set it must point to bin folder.
"""
    __slots__ = ()

    def envNames(self):
        return ['gccBin', 'gccPrefix', 'gccPostfix']

    def _installTool(self, env):
        self._install.deb(['gcc'])
        self._install.rpm(['gcc'])
        self._install.emerge(['sys-devel/gcc'])
        self._install.pacman(['gcc'])
        self._install.apk(['gcc'])

    def initEnv(self, env):
        gcc_bin = env.get('gccBin', None)

        pref = env.setdefault('gccPrefix', '')
        postf = env.setdefault('gccPostfix', '')

        if gcc_bin:
            if self._ospath.exists(gcc_bin):
                self._have_tool = True
        else:
            gcc = pref + 'gcc' + postf
            gcc = self._pathutil.which(gcc)

            if gcc:
                env['gccBin'] = gcc
                self._have_tool = True



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class sshTool(RunEnvTool):
    """Secure Shell client.

Home: https://www.openssh.com/

Note:
    * sshStrictHostKeyChecking is set to "yes" by default.
      It is used by tools depending on ssh.
"""
    __slots__ = ()

    def envNames(self):
        return ['sshBin', 'sshStrictHostKeyChecking']

    def _installTool(self, env):
        self._install.deb(['openssh-client'])
        self._install.rpm(['openssh'])
        self._install.emerge(['virtual/ssh'])
        self._install.pacman(['openssh'])
        self._install.apk('openssh-client')

    def initEnv(self, env):
        env.setdefault('sshStrictHostKeyChecking', 'yes')
        super(sshTool, self).initEnv(env)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .gemtoolmixin import GemToolMixIn


class puppetTool(GemToolMixIn, BuildTool):
    """Puppet system automation.

Home: https://puppet.com/

Primary purpose is to support Puppet module development.
"""
    __slots__ = ()

    METADATA_FILE = 'metadata.json'

    def envNames(self):
        return ['puppetVer', 'puppetBin']

    def initEnv(self, env):
        super(puppetTool, self).initEnv(env)
        puppet_ver = env.get('puppetVer', None)

        if self._have_tool and puppet_ver:
            try:
                found_ver = self._executil.callExternal(
                    [env['puppetBin'], '--version'], verbose=False)
                self._have_tool = found_ver.find(puppet_ver) >= 0
            except:
                self._have_tool = False
                del env['puppetBin']

    def autoDetectFiles(self):
        return self.METADATA_FILE

    def loadConfig(self, config):
        content = self._pathutil.loadJSONConfig(self.METADATA_FILE)
        if content is None:
            return

        for f in ('name', 'version'):
            if f in content and f not in config:
                config[f] = content[f]

        config['package'] = ['pkg']

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self.METADATA_FILE, updater)

    def onPrepare(self, config):
        self._pathutil.rmTree('pkg')

    def onBuild(self, config):
        puppetBin = config['env']['puppetBin']
        cmd = [puppetBin, 'module', 'build']
        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        package_file = 'pkg/{0}-{1}.tar.gz'.format(
            config['name'],
            config['version']
        )

        if not self._ospath.exists(package_file):
            self._errorExit(
                'Puppet Module built package is missing: ' + package_file)

        self._pathutil.addPackageFiles(config, package_file)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import print_function

from ..coloring import Coloring
from ..mixins.ondemand import ext as _ext


def info(msg, label=None):
    if label:  # for backward compatibility
        infoLabel(label, msg)
        return

    line = Coloring.info('INFO: ' + msg)
    print(line, file=_ext.sys.stderr)


def infoLabel(label, msg):
    line = Coloring.infoLabel(label) + Coloring.info(msg)
    print(line, file=_ext.sys.stderr)


def warn(msg):
    print(Coloring.warn('WARNING: ' + msg), file=_ext.sys.stderr)


def errorExit(msg):
    raise RuntimeError(msg)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from .bashtoolmixin import BashToolMixIn
from .javatoolmixin import JavaToolMixIn


class SdkmanToolMixIn(BashToolMixIn, JavaToolMixIn):
    __slots__ = ()

    def _needJDK(self):
        return True

    def getDeps(self):
        return (
            super(SdkmanToolMixIn, self).getDeps() +
            ['sdkman', 'java'] +
            (self._needJDK() and ['jdk'] or [])
        )

    def _sdkName(self):
        return self._name

    def _binName(self):
        return self._name

    def _setAutoAnswer(self, env):
        self._callBash(env,
                       'grep -q "sdkman_auto_answer=true" {0} || echo "sdkman_auto_answer=true" >> {0}'
                       .format(self._pathutil.safeJoin(env['sdkmanDir'], 'etc', 'config')),
                       verbose=False
                       )

    def _callSdkman(self, env, cmd, verbose=True):
        return self._callBash(env,
                              'source {0} >/dev/null && sdk {1}'.format(
                                  env['sdkmanInit'],
                                  cmd
                              ),
                              verbose=verbose
                              )

    def _javaVersion(self, env):
        if not env.get('javaBin', None):
            return 0

        java_ver = self._callBash(
            env, '{0} -version 2>&1'.format(env['javaBin']), verbose=False)
        return int(self._ext.re.search('version "1\.([0-9]+)\.', java_ver).group(1))

    def _installTool(self, env):
        self._setAutoAnswer(env)
        self._callSdkman(env,
                         'install {0} {1}'.format(
                             self._sdkName(),
                             env.get(self._name + 'Ver', '')
                         )
                         )

    def _updateTool(self, env):
        self._setAutoAnswer(env)
        self._callSdkman(env,
                         'upgrade {0}'.format(
                             self._sdkName()
                         )
                         )

    def uninstallTool(self, env):
        ospath = self._ospath
        tool_dir = ospath.join(
            env['sdkmanDir'], 'candidates', self._sdkName())
        self._pathutil.rmTree(tool_dir)

    def initEnv(self, env):
        ospath = self._ospath
        if not env.get('sdkmanDir', None):
            return

        tool_dir = ospath.join(
            env['sdkmanDir'], 'candidates', self._sdkName(), 'current')

        if not ospath.exists(tool_dir):
            return

        try:
            env_to_set = self._callSdkman(env,
                                          'use {0} {1} >/dev/null && env | grep -i {0}'
                                          .format(self._sdkName(), env.get(self._name + 'Ver', '')),
                                          verbose=False
                                          )
        except:
            return

        if env_to_set:
            self._pathutil.updateEnvFromOutput(env_to_set)
            super(SdkmanToolMixIn, self).initEnv(env, self._binName())



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool


class exeTool(RuntimeTool):
    """Dummy tool to execute files directly"""
    __slots__ = ()

    def initEnv(self, env):
        self._have_tool = True

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([svc['path']] + args)



#!/usr/bin/env python

#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""FutoIn CID v{0}

This command is alias for "cid tool exec" of FutoIn Continuous Integration & Delivery Tool.

Experience has shown that this command is used a lot in development process and
requires a dedicated alias.

Usage:
    cte <tool_name> [<tool_arg>...]
    cte list
"""

from __future__ import print_function

from .cidtool import CIDTool
from .coloring import Coloring

import os
import sys

if sys.version_info < (2, 7):
    print('Sorry, but only Python version >= 2.7 is supported!', file=sys.stderr)
    sys.exit(1)

from . import __version__ as version
__all__ = ['run']


def runInner():
    argv = sys.argv

    try:
        tool = argv[1]
    except IndexError:
        tool = None

    if tool is None or tool in ('-h', '--help', 'help'):
        print(__doc__.format(version))
    else:
        ospath = os.path

        if 'CID_COLOR' in os.environ:
            Coloring.enable(os.environ['CID_COLOR'] == 'yes')

        overrides = {}
        overrides['wcDir'] = ospath.realpath('.')

        if tool not in ('list',):
            overrides['tool'] = tool

        #---
        overrides['toolVer'] = None
        overrides['toolDetect'] = False

        #---
        cit = CIDTool(overrides=overrides)

        if tool == 'list':
            cit.tool_list()
        else:
            cit.tool_exec(tool, argv[2:])


def run():
    try:
        runInner()
    except Exception as e:
        print(file=sys.stderr)
        print(Coloring.error('ERROR: ' + str(e)), file=sys.stderr)
        print(file=sys.stderr)
        import traceback
        print(Coloring.warn(traceback.format_exc()), file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt as e:
        print(file=sys.stderr)
        print(Coloring.error('Exit on user abort'), file=sys.stderr)
        print(file=sys.stderr)
        sys.exit(1)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import log as _log

_BUIID_DEP_PREFIX = '_bdep_'
_SYSTEM_VER = 'system'


def essential():
    _ext.install.SUSEConnectVerArch('sle-module-toolchain')
    _ext.install.SUSEConnectVerArch('sle-sdk', True)

    _ext.install.deb([
        'build-essential',
    ])
    _ext.install.rpm([
        'binutils',
        'gcc',
        'gcc-c++',
        'glibc-devel',
        'libtool',
        'make',
        'patch'
    ])

    _ext.install.yum('redhat-rpm-config')

    _ext.install.apk([
        'build-base',
        'linux-headers',
    ])


def require(env, dep):
    dep = _ext.configutil.listify(dep)

    essential()

    for d in dep:
        m = '{0}{1}'
        m = m.format(_BUIID_DEP_PREFIX, d.replace('-', ''))
        m = globals().get(m, None)

        if m:
            m(env)
        else:
            _log.errorExit('Unknown build dep "{0}"'.format(d))


def available():
    ret = []

    for m in globals():
        if m.startswith(_BUIID_DEP_PREFIX):
            m = m.replace(_BUIID_DEP_PREFIX, '')
            ret.append(m)

    return ret


def _bdep_ruby(env):
    detect = _ext.detect
    ver = env['rubyVer']

    if ver == _SYSTEM_VER:
        _ext.install.deb(['ruby-dev'])
        _ext.install.rpm(['ruby-devel'])
        _ext.install.apk(['ruby-dev'])
    elif detect.isSCLSupported():
        devver = ver.replace('.', '')

        if devver == '19':
            sclname = 'ruby193-ruby-devel'
        elif devver == '20':
            sclname = 'ruby200-ruby-devel'
        else:
            sclname = 'rh-ruby{0}-ruby-devel'.format(devver)

        _ext.install.rpm(sclname)
    elif detect.isDebian() or detect.isUbuntu():
        if ver == '1.9':
            pkgver = '1.9.[0-9]'
        else:
            pkgver = ver

        _ext.install.deb([
            'ruby{0}-dev'.format(pkgver),
        ])


def _bdep_python(env):
    if int(env['pythonVer'].split('.')[0]) == 3:
        _ext.install.deb(['python3-dev'])
        _ext.install.zypper(['python3-devel'])
        _ext.install.yum(['python3-devel'])
        _ext.install.yum(['python34-devel'])
        _ext.install.apk(['python3-dev'])
    else:
        _ext.install.deb(['python-dev'])
        _ext.install.rpm(['python-devel'])
        _ext.install.apk(['python2-dev'])


def _bdep_ssl(env):
    _bdep_ssl10(env)


def _bdep_ssl10(env):
    apt_cache = _ext.pathutil.which('apt-cache')

    if apt_cache and _ext.executil.callExternal([apt_cache, 'search', 'libssl1.0-dev']).strip():
        _ext.install.deb('libssl1.0-dev')
    else:
        _ext.install.deb('libssl-dev')
    _ext.install.rpm('openssl-devel')
    _ext.install.rpm('libopenssl-devel')
    _ext.install.apk('libressl-dev')
    #_ext.install.pacman('openssl')
    _ext.install.pacman('openssl-1.0')
    _ext.install.brew('openssl')


def _bdep_mysqlclient(env):
    _ext.install.deb('libmysqlclient-dev')
    _ext.install.deb('default-libmysqlclient-dev')

    if _ext.detect.isOracleLinux():
        _ext.install.yum(['mariadb-devel', 'mariadb-libs'])
    else:
        _ext.install.yum('mysql-devel')

    _ext.install.zypper('libmysqlclient-devel')
    _ext.install.pacman('libmariadbclient')
    _ext.install.apk('mariadb-dev')
    _ext.install.brew('mysql')


def _bdep_postgresql(env):
    _ext.install.deb('libpq-dev')
    _ext.install.rpm(['postgresql-devel', 'postgresql-libs'])
    _ext.install.pacman('postgresql')
    _ext.install.apk('postgresql-dev')
    _ext.install.brew('postgresql')


def _bdep_imagemagick(env):
    _ext.install.deb('libmagick-dev')
    _ext.install.rpm(['imagemagick', 'imagemagick-devel'])
    _ext.install.pacman('imagemagick')
    _ext.install.apk('imagemagick-dev')
    _ext.install.brew('imagemagick')


def _bdep_tzdata(env):
    _ext.install.deb('tzdata')
    _ext.install.rpm('tzdata')
    _ext.install.pacman('tzdata')
    _ext.install.apk('tzdata')


def _bdep_libxml2(env):
    _ext.install.deb('libxml2-dev')
    _ext.install.rpm('libxml2-devel')
    _ext.install.pacman('libxml2')
    _ext.install.apk('libxml2')
    _ext.install.brew('libxml2')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .curltoolmixin import CurlToolMixIn


class composerTool(CurlToolMixIn, BuildTool):
    """Dependency Manager for PHP.

Home: https://getcomposer.org/

Composer is installed in composerDir as single Phar with "composer" name without
extension.
composerDir is equal to user's ~/bin/ folder by default.
"""
    __slots__ = ()

    COMPOSER_JSON = 'composer.json'

    def _installTool(self, env):
        composer_dir = env['composerDir']
        php_bin = env['phpBin']
        composer_get = env.get(
            'composerGet', 'https://getcomposer.org/installer')

        composer_installer = self._callCurl(
            env, [composer_get], binary_output=True)

        if not self._ospath.exists(composer_dir):
            self._os.makedirs(composer_dir)

        self._phputil.installExtensions(env, [
            'phar',
            'openssl',
            'json',
            'filter',
            'hash',
            'mbstring',
        ])
        self._phputil.installExtensions(env, [
            'zip',
            'apcu',
            'zlib',
        ], True)

        self._executil.callExternal(
            [
                php_bin, '--',
                '--install-dir=' + composer_dir,
                '--filename=composer'
            ],
            input=composer_installer,
            binary_input=True)

    def _updateTool(self, env):
        self._executil.callExternal([env['composerBin'], 'self-update'])

    def uninstallTool(self, env):
        self._os.remove(env['composerBin'])
        self._have_tool = False

    def envNames(self):
        return ['composerDir', 'composerBin', 'composerGet']

    def initEnv(self, env):
        ospath = self._ospath
        bin_dir = env['binDir']
        composer_dir = env.setdefault('composerDir', bin_dir)
        composer_bin = env.setdefault(
            'composerBin', ospath.join(composer_dir, 'composer'))

        self._pathutil.addBinPath(composer_dir)

        self._have_tool = ospath.exists(composer_bin)

    def autoDetectFiles(self):
        return self.COMPOSER_JSON

    def getDeps(self):
        return ['php', 'git'] + CurlToolMixIn.getDeps(self)

    def loadConfig(self, config):
        content = self._pathutil.loadJSONConfig(self.COMPOSER_JSON)
        if content is None:
            return

        for f in ('name', 'version'):
            if f in content and f not in config:
                config[f] = content[f]

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self.COMPOSER_JSON, updater, indent=4)

    def onPrepare(self, config):
        composerBin = config['env']['composerBin']
        cmd = [composerBin, 'install']
        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        composerBin = config['env']['composerBin']
        cmd = [composerBin, 'install', '--no-dev', '--no-scripts']
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn
from .rvmtool import rvmTool


class rubyTool(BashToolMixIn, RuntimeTool):
    """Ruby is a dynamic, open source programming language.

Home: https://www.ruby-lang.org/en/

By default the latest available Ruby binary is used for the following OSes:
* Debian & Ubuntu - uses Brightbox builds 1.9, 2.0, 2.1, 2.2, 2.3, 2.4.
* CentOS, RHEL & Oracle Linux - uses SCL 1.9, 2.0, 2.2, 2.3

You can forbid source builds by setting rubyBinOnly to non-empty string.

Otherwise, System Ruby is used by default.

If rubyVer is set then RVM is used to setup custom rubies.
That may lead to long time and resource consumption due to compilation,
if binary versions are not found for specific system.

Note: RUBY_ENV and RAILS_ENV are set based on rubyEnv or .env.type
"""
    __slots__ = ()

    def getDeps(self):
        return ['rvm']

    def _installTool(self, env):
        rvmTool('rvm').ensureGpgKeys(env)

        if env['rubyFoundBinary']:
            self._installBinaries(env)
            return

        self._buildDeps(env)
        self._executil.callExternal([
            env['rvmBin'], 'install', env['rubySourceVer'], '--autolibs=read-only'
        ])
        self._executil.callExternal([
            env['rvmBin'], 'cleanup', 'all'
        ])

    def _installBinaries(self, env):
        detect = self._detect
        install = self._install
        pathutil = self._pathutil
        executil = self._executil

        ver = env['rubyVer']
        rvm_ruby_ver = 'system-{0}'.format(ver)
        pkgver = ver

        if detect.isDebian() or detect.isUbuntu():
            code_name = self._detect.osCodeName()

            if code_name in ['stretch', 'buster', 'sid', 'testing']:
                install.aptRepo('jessie-ssl10bp',
                                'deb http://deb.debian.org/debian jessie-backports main')

            repo = env['rubyBrightboxRepo']

            install.aptRepo(
                'brightbox-ruby',
                "deb {0} $codename$ main".format(repo),
                self._GPG_BIRGHTBOX_REPO,
                codename_map={
                    # Ubuntu
                    # Debian
                    'jessie': 'trusty',
                    'stretch': 'xenial',
                    'testing': 'xenial',
                    'sid': 'xenial',
                },
                repo_base='{0}/dists'.format(repo)
            )

            if ver == '1.9':
                pkgver = '1.9.[0-9]'


            if detect.isDebian():
                UBUNTU_MIRROR = 'https://debian.charite.de/ubuntu'
                UBUNTU_MIRROR += '/pool/main/r/readline6/libreadline6_6.3-8ubuntu2_amd64.deb'
                install.dpkg(env, 'libreadline6', UBUNTU_MIRROR)

            install.deb([
                'ruby{0}'.format(pkgver),
                'ruby{0}-dev'.format(pkgver),
            ])
            ruby_bins = self._ext.glob.glob('/usr/bin/ruby{0}*'.format(ver))

            if len(ruby_bins) == 0:
                self._errorExit('No Ruby found for version {0}'.format(ver))

            ruby_bin = ruby_bins[0]

        elif detect.isSCLSupported():
            sclname = self._rubySCLName(ver)

            install.yumSCL()

            install.yum(sclname)

            # required for LD_LIBRARY_PATH
            env_to_set = executil.callExternal(
                ['scl', 'enable', sclname, 'env'], verbose=False)
            pathutil.updateEnvFromOutput(env_to_set)

            ruby_bin = executil.callExternal(
                ['scl', 'enable', sclname, 'which ruby'], verbose=False).strip()

        elif detect.isMacOS():
            formula = 'ruby@{0}'.format(ver)
            install.brew(formula)
            return
        else:
            self._systemDeps()
            rvm_ruby_ver = 'system'
            ruby_bin = pathutil.which('ruby')

        executil.callExternal([
            env['rvmBin'], 'remove', 'ext-{0}'.format(rvm_ruby_ver)
        ], suppress_fail=True)

        executil.callExternal([
            env['rvmBin'], 'mount', ruby_bin, '-n', rvm_ruby_ver
        ])

    def _rubySCLName(self, ver):
        pkgver = ver.replace('.', '')

        if pkgver == '19':
            sclname = 'ruby193'
        elif pkgver == '20':
            sclname = 'ruby200'
        else:
            sclname = 'rh-ruby' + pkgver

        return sclname

    def _fixRvmLinks(self, env, name, ver):
        ospath = self._ospath
        os = self._os
        glob = self._ext.glob
        bin_dir = ospath.join(env['rvmDir'], 'rubies', name, 'bin')

        for f in ['erb', 'gem', 'irb', 'rake', 'rdoc', 'ri', 'ruby', 'testrb']:
            f = ospath.join(bin_dir, f)
            res = glob.glob('{0}{1}*'.format(f, ver))

            if res:
                res = ospath.basename(res[0])

                if os.readlink(f) == res:
                    continue

                try:
                    os.unlink(f)
                except Exception as e:
                    self._warn(str(e))

                os.symlink(res, f)

    def _updateTool(self, env):
        if not env['rubyFoundBinary']:
            self._installTool(env)

    def uninstallTool(self, env):
        ruby_ver = env['rubyVer']

        if not env['rubyFoundBinary']:
            self._executil.callExternal([
                env['rvmBin'], 'uninstall', env['rubyVer']
            ])
            self._have_tool = False

    def envNames(self):
        return ['rubyVer', 'rubyBin', 'rubyBinOnly', 'rubyForceBuild', 'rubySourceVer', 'rubyEnv']

    def initEnv(self, env):
        environ = self._environ
        ospath = self._ospath
        detect = self._detect
        path = self._pathutil

        #---
        ruby_env = env.get('rubyEnv', '')

        if ruby_env:
            pass
        elif env['type'] == 'dev':
            ruby_env = 'development'
        else:
            ruby_env = 'production'

        environ['RUBY_ENV'] = ruby_env
        environ['RAILS_ENV'] = ruby_env

        #---
        if 'GEM_HOME' in environ:
            path.delEnvPath('PATH', environ['GEM_HOME'])
            path.delEnvPath('GEM_PATH', environ['GEM_HOME'])
            del environ['GEM_HOME']

        #---
        rubyForceBuild = env.setdefault('rubyForceBuild', False)
        rubyBinOnly = env.setdefault('rubyBinOnly', not rubyForceBuild)

        if rubyBinOnly and rubyForceBuild:
            self._warn('"rubyBinOnly" and "rubyForceBuild" do not make sense'
                       ' when set together!')

        #---
        if detect.isDebian() or detect.isUbuntu():
            bb_repo = 'http://ppa.launchpad.net/brightbox/ruby-ng/ubuntu'
            ruby_binaries = ['1.9', '2.0', '2.1', '2.2', '2.3', '2.4', '2.5', '2.6']

            code_name = self._detect.osCodeName()

            if code_name in []:
                # 1.9 build is broken on LaunchPad
                bb_repo = 'http://ppa.launchpad.net/brightbox/ruby-ng-experimental/ubuntu'

            env.setdefault('rubyBrightboxRepo', bb_repo)

        elif detect.isSCLSupported():
            if detect.isCentOS():
                ruby_binaries = ['1.9', '2.0', '2.2', '2.3', '2.4', '2.5']
            else:
                ruby_binaries = ['1.9', '2.0', '2.2', '2.3', '2.4', '2.5']
        elif detect.isMacOS():
            ruby_binaries = ['1.8', '1.9', '2.0', '2.2', '2.3', '2.4', '2.5', '2.6']
        else:
            ruby_binaries = None

        #---
        if ruby_binaries and not rubyForceBuild:
            ruby_ver = env.setdefault('rubyVer', ruby_binaries[-1])
            foundBinary = ruby_ver in ruby_binaries

            rvm_ruby_ver = ruby_ver

            if foundBinary:
                rvm_ruby_ver = 'ext-system-{0}'.format(ruby_ver)

                # required for LD_LIBRARY_PATH
                if detect.isSCLSupported():
                    sclname = self._rubySCLName(ruby_ver)

                    try:
                        env_to_set = self._executil.callExternal(
                            ['scl', 'enable', sclname, 'env'], verbose=False)
                        self._pathutil.updateEnvFromOutput(env_to_set)
                    except self._ext.subprocess.CalledProcessError:
                        pass
                    except OSError:
                        pass
                elif detect.isMacOS():
                    env['rubyFoundBinary'] = True
                    formula = 'ruby@{0}'.format(ruby_ver)
                    brew_prefix = env['brewDir']
                    ruby_bin_dir = ospath.join(
                        brew_prefix, 'opt', formula, 'bin')

                    if ospath.exists(ruby_bin_dir):
                        self._pathutil.addBinPath(ruby_bin_dir, True)
                        super(rubyTool, self).initEnv(env)
                        self._environ['rubyVer'] = ruby_ver
                    return
                elif detect.isDebian() or detect.isUbuntu():
                    self._fixRvmLinks(env, rvm_ruby_ver, ruby_ver)
        else:
            ruby_ver = env.setdefault('rubyVer', self.SYSTEM_VER)
            foundBinary = ruby_ver == self.SYSTEM_VER
            rvm_ruby_ver = foundBinary and ruby_ver or 'ext-system'

        #---
        rvm_dir = env['rvmDir']
        env['rubyFoundBinary'] = foundBinary

        if rubyForceBuild or not foundBinary:
            rvm_ruby_ver = env.setdefault('rubySourceVer', ruby_ver or 'ruby')

        try:
            env_to_set = self._callBash(env,
                                        'source {0} && \
                rvm use {1} >/dev/null && \
                env | grep "rvm"'.format(env['rvmInit'], rvm_ruby_ver),
                                        verbose=False
                                        )
        except:
            return

        if env_to_set:
            self._pathutil.updateEnvFromOutput(env_to_set)
            super(rubyTool, self).initEnv(env)
            self._environ['rubyVer'] = ruby_ver

    def _buildDeps(self, env):
        self._builddep.require(env, 'ssl')

        # APT
        #---
        self._install.deb([
            'build-essential',
            'gawk',
            'make',
            'libc6-dev',
            'zlib1g-dev',
            'libyaml-dev',
            'libsqlite3-dev',
            'sqlite3',
            'autoconf',
            'libgmp-dev',
            'libgdbm-dev',
            'libncurses5-dev',
            'automake',
            'libtool',
            'bison',
            'pkg-config',
            'libffi-dev',
            'libgmp-dev',
            'libreadline-dev',
        ])

        # Extra repo before the rest
        #---
        self._install.yumEPEL()

        self._install.rpm([
            'binutils',
            'patch',
            'libyaml-devel',
            'autoconf',
            'gcc',
            'gcc-c++',
            'glibc-devel',
            'readline-devel',
            'zlib-devel',
            'libffi-devel',
            'automake',
            'libtool',
            'bison',
            'sqlite-devel',
            'make',
            'm4',
            'gdbm-devel',
            'sqlite3-devel',
        ])

        #---
        self._install.emergeDepsOnly(['dev-lang/ruby'])
        self._install.pacman(['ruby'])
        self._install.apk('build-base')

    def _systemDeps(self):
        self._install.debrpm(['ruby'])
        self._install.emerge(['dev-lang/ruby'])
        self._install.pacman(['ruby'])
        self._install.apk(['ruby',
                           'ruby-bigdecimal',
                           'ruby-libs',
                           'ruby-io-console',
                           'ruby-irb',
                           'ruby-json',
                           'ruby-minitest',
                           'ruby-net-telnet',
                           'ruby-power_assert',
                           'ruby-xmlrpc'
                           ])

    _GPG_BIRGHTBOX_REPO = '''
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: SKS 1.1.6
Comment: Hostname: keyserver.ubuntu.com

mI0ETKTCMQEEAMX3ttL4YFO5AQ7Z6L5gaGw57CJBQl6jCv6lka0p8DaGNkeX0Rs9DhINa8qR
hxJCPK6ijeoNss69G/ni+sMSRViJBFWXzitEE1ew5YM2sw7wLE3guToDu60kaDwIn5mR3GTx
cgqDrQeCuGZJgz3e2lgmGYw2rAhMe78rRgkR5GFvABEBAAG0G0xhdW5jaHBhZCBQUEEgZm9y
IEJyaWdodGJveIi4BBMBAgAiBQJMpMIxAhsDBgsJCAcDAgYVCAIJCgsEFgIDAQIeAQIXgAAK
CRD12l8Jwxc6pl2BA/4p5DFEpGVvkgLj7/YLYCtYmZDw8i/drGbkWfIQiOgPWIf8QgpJXVME
1tkH8N1ssjbJlUKl/HubNBKZ6HDyQsQASFug+eI6KhSFMScDBf/oMX3zVCTTvUkgJtOWYc5d
77zJacEUGoSEx63QUJVvp/LAnqkZbt17JJL6HOou/CNicw==
=G8vE
-----END PGP PUBLIC KEY BLOCK-----
'''

    def tuneDefaults(self, env):
        return {
            'minMemory': '8M',
            'socketType': 'none',
            'scalable': False,
            'reloadable': False,
            'multiCore': True,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext

_yum = _ext.pathutil.which('yum')
_dnf = _ext.pathutil.which('dnf')
_zypper = _ext.pathutil.which('zypper')
_rpm = _ext.pathutil.which('rpm')


def yum(packages):
    if not _ext.detect.isYum() and not _ext.detect.isDnf():
        return

    yum = _dnf or _yum

    packages = _ext.configutil.listify(packages)

    _ext.executil.trySudoCall(
        [yum, 'install', '-y'] + packages,
        errmsg='you may need to install the packages manually !'
    )


def zypper(packages):
    if not _ext.detect.isZypper():
        return

    packages = _ext.configutil.listify(packages)

    _ext.executil.trySudoCall(
        [_zypper, '--non-interactive', 'install', '-y'] + packages,
        errmsg='you may need to install the packages manually !'
    )


def rpm(packages):
    yum(packages)
    zypper(packages)


def rpmKey(gpg_key):
    if not gpg_key:
        return

    if not _ext.detect.isRPM():
        return

    os = _ext.os

    tmp_dir = _ext.pathutil.tmpCacheDir(prefix='cidgpg')
    tf = os.path.join(tmp_dir, 'key.gpg')
    _ext.pathutil.writeBinaryFile(tf, gpg_key)

    _ext.executil.trySudoCall(
        [_rpm, '--import', tf],
        errmsg='you may need to import the PGP key manually!'
    )

    os.remove(tf)


def yumRepo(name, url, gpg_key=None, releasevermax=None, repo_url=False):
    if not _ext.detect.isYum() and not _ext.detect.isDnf():
        return

    rpmKey(gpg_key)

    repo_file = None
    repo_info = None
    pathutil = _ext.pathutil

    #---
    if repo_url:
        tmp_dir = pathutil.tmpCacheDir(prefix='cidrepo')
        repo_file = '{0}.repo'.format(name)
        repo_file = _ext.ospath.join(tmp_dir, repo_file)

        repo_info = "\n".join([
            '[nginx]',
            'name={0} repo'.format(name),
            'baseurl={0}'.format(url),
            'gpgcheck=1',
            'enabled=1',
            ''
        ])

        pathutil.writeTextFile(repo_file, repo_info)
        url = repo_file

    #---
    try:
        dist_ver = _ext.detect.linuxDistMajorVer()

        if releasevermax is not None and int(dist_ver) > releasevermax:
            dist_ver = releasevermax

            if not repo_info:
                repo_info = _ext.urllib.urlopen(url).read()
                repo_info = _ext.executil.toString(repo_info)

            repo_info = repo_info.replace('$releasever', str(dist_ver))

            if not repo_file:
                tmp_dir = _ext.pathutil.tmpCacheDir(prefix='cidrepo')
                repo_file = url.split('/')[-1]
                repo_file = _ext.ospath.join(tmp_dir, repo_file)

            pathutil.writeTextFile(repo_file, repo_info)
            url = repo_file

        #---
        if _dnf:
            yum(['dnf-plugins-core'])
            _ext.executil.trySudoCall(
                [_dnf, 'config-manager', '--add-repo', url],
                errmsg='you may need to add the repo manually!'
            )

        else:
            yum(['yum-utils'])
            yumcfgmgr = _ext.pathutil.which('yum-config-manager')
            _ext.executil.trySudoCall(
                [yumcfgmgr, '--add-repo', url],
                errmsg='you may need to add the repo manually!'
            )
    finally:
        if repo_file:
            _ext.pathutil.rmTree(repo_file)


def zypperRepo(name, url, gpg_key=None, yum=False):
    if not _ext.detect.isZypper():
        return

    rpmKey(gpg_key)

    if yum:
        cmd = [_zypper, '--non-interactive', 'addrepo', '-t', 'YUM', url, name]
    elif name:
        cmd = [_zypper, '--non-interactive', 'addrepo', url, name]
    else:
        cmd = [_zypper, '--non-interactive', 'addrepo', url]

    _ext.executil.trySudoCall(
        cmd,
        errmsg='you may need to add the repo manually!'
    )

    if name:
        _ext.executil.trySudoCall(
            [_zypper, '--non-interactive', 'refresh', '-r', name],
            errmsg='you may need to add the repo manually!'
        )
    else:
        _ext.executil.trySudoCall(
            [_zypper, '--non-interactive', 'refresh'],
            errmsg='you may need to add the repo manually!'
        )


def yumEnable(repo):
    if _ext.detect.isDnf():
        yum(['dnf-plugins-core'])
        _ext.executil.trySudoCall(
            [_dnf, 'config-manager', '--enable', repo],
            errmsg='you may need to add the repo manually!'
        )

    elif _ext.detect.isYum():
        yum(['yum-utils'])

        yumcfgmgr = _ext.pathutil.which('yum-config-manager')

        _ext.executil.trySudoCall(
            [yumcfgmgr, '--enable', repo],
            errmsg='You may need to enable the repo manually'
        )


def yumEPEL():
    detect = _ext.detect

    if not detect.isRPM():
        return

    if detect.isCentOS():
        yum(['epel-release'])
    else:
        ver = detect.linuxDistMajorVer()
        yum(
            ['https://dl.fedoraproject.org/pub/epel/epel-release-latest-{0}.noarch.rpm'.format(ver)])


def yumSCL():
    detect = _ext.detect

    if not detect.isSCLSupported():
        return

    ver = detect.linuxDistMajorVer()

    if detect.isRHEL():
        yumEnable('rhel-server-rhscl-{0}-rpms'.format(ver))
    elif detect.isCentOS():
        yum('centos-release-scl-rh')
    elif detect.isOracleLinux():
        yumOLPublic([
            'software_collections',
            'latest',
            'optional_latest',
        ])

    yum('scl-utils')


def yumIUS():
    detect = _ext.detect

    if not detect.isSCLSupported():
        return

    yumEPEL()
    ver = detect.linuxDistMajorVer()

    if detect.isCentOS():
        yum('https://centos{0}.iuscommunity.org/ius-release.rpm'.format(ver))
    else:
        yum('https://rhel{0}.iuscommunity.org/ius-release.rpm'.format(ver))


def yumRHELRepo(repos):
    detect = _ext.detect

    if not detect.isRHEL():
        return

    ver = detect.linuxDistMajorVer()
    repos = _ext.configutil.listify(repos)

    scrmgr = _ext.pathutil.which('subscription-manager')

    for r in repos:
        _ext.executil.trySudoCall(
            [scrmgr, 'repos', '--enable', 'rhel-{0}-{1}'.format(ver, r)],
            errmsg='you may need to enable the repo manually!'
        )


def yumOLPublic(repos):
    detect = _ext.detect

    if not detect.isOracleLinux():
        return

    ver = detect.linuxDistMajorVer()
    repos = _ext.configutil.listify(repos)

    if not _ext.ospath.exists('/etc/yum.repos.d/public-yum-ol{0}.repo'.format(ver)):
        yumRepo('public-yum-o1{0}'.format(ver),
                'http://yum.oracle.com/public-yum-ol{0}.repo'.format(ver))

    for r in repos:
        yumEnable('ol{0}_{1}'.format(ver, r))


def SUSEConnect(ext):
    if not _ext.detect.isSLES():
        return

    _ext.executil.trySudoCall(
        ['/usr/bin/SUSEConnect', '--product', ext, '-r', ''],
        errmsg='you may need to add the extension manually!'
    )


def SUSEConnectVerArch(ext, full_ver=False):
    detect = _ext.detect

    if not detect.isSLES():
        return

    if full_ver:
        sles_ver = detect.osRelease()
    else:
        sles_ver = detect.linuxDistMajorVer()

    base_arch = detect.basearch()
    SUSEConnect('{0}/{1}/{2}'.format(ext, sles_ver, base_arch))



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import log as _log
from . import complex_memo as _complex_memo


def which(program):
    return cachedWhich(program, _ext.os.environ["PATH"])


@_complex_memo
def cachedWhich(program, env_path):
    "Copied from stackoverflow"

    os = _ext.os
    ospath = _ext.ospath

    def is_exe(fpath):
        return ospath.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = ospath.split(program)

    if fpath:
        if is_exe(program):
            return program
    else:
        for path in env_path.split(os.pathsep):
            path = path.strip('"')
            exe_file = safeJoin(path, program)

            if is_exe(exe_file):
                return exe_file

    return None


def safeJoin(*args):
    ospath = _ext.ospath
    paths = list(args[:1]) + [p.lstrip(ospath.sep) or '.' for p in args[1:]]
    return ospath.join(*paths)


def addEnvPath(env_name, add_dir, first=False):
    os = _ext.os
    environ = os.environ

    if env_name in os.environ:
        dir_list = os.environ[env_name].split(os.pathsep)
    else:
        dir_list = []

    if add_dir not in dir_list:
        if first:
            dir_list[0:0] = [add_dir]
        else:
            dir_list.append(add_dir)

        os.environ[env_name] = os.pathsep.join(dir_list)


def delEnvPath(env_name, del_dir, fail=False):
    os = _ext.os
    environ = os.environ

    if env_name in environ:
        dir_list = environ[env_name].split(os.pathsep)

        try:
            del dir_list[dir_list.index(del_dir)]
        except ValueError:
            if fail:
                raise
    elif fail:
        # trigger KeyError
        return environ[env_name]


def addBinPath(bin_dir, first=False):
    addEnvPath('PATH', bin_dir, first=first)


def addPackageFiles(config, pattern):
    files = _ext.glob.glob(pattern)

    if not files:
        _log.errorExit(
            'Failed to find created packages of "{0}" pattern'.format(pattern))

    config.setdefault('packageFiles', [])
    config['packageFiles'] += files


def updateEnvFromOutput(env_to_set):
    environ = _ext.os.environ
    env_to_set = env_to_set.split("\n")

    for e in env_to_set:
        if not e:
            continue
        n, v = e.split('=', 1)
        environ[n] = v


def userHome():
    return _ext.os.environ['HOME']


def deployHome():
    environ = _ext.os.environ
    return environ.get('CID_DEPLOY_HOME', environ['HOME'])


def loadJSONConfig(file_name, defvalue=None):
    if _ext.ospath.exists(file_name):
        with open(file_name, 'r') as content_file:
            content = content_file.read()
            return _ext.configutil.parseJSON(content)
    else:
        return defvalue


def updateJSONConfig(file_name, updater, indent=2, separators=(',', ': ')):
    content = loadJSONConfig(file_name)

    if content is None:
        return []

    updater(content)

    writeJSONConfig(file_name, content, indent, separators)

    return [file_name]


def writeJSONConfig(file_name, content, indent=2, separators=(',', ': ')):
    with open(file_name, 'w') as content_file:
        content = _ext.json.dumps(
            content, indent=indent, separators=separators)
        content_file.write(content)
        content_file.write("\n")

#---


def readTextFile(file_name):
    with open(file_name, 'r') as content_file:
        return content_file.read()


def updateTextFile(file_name, updater):
    content = readTextFile(file_name)
    content = updater(content)
    writeTextFile(file_name, content)

    return [file_name]


def writeTextFile(file_name, content):
    with open(file_name, 'w') as content_file:
        content_file.write(content)


def writeBinaryFile(file_name, content):
    content = _ext.executil.toBytes(content)

    with open(file_name, 'wb') as content_file:
        content_file.write(content)


def writeIni(file_name, content):
    str_content = []
    listify = _ext.configutil.listify

    for (sn, sv) in content.items():
        str_content.append('[{0}]'.format(sn))

        for (cn, cv) in sv.items():
            for cvi in listify(cv):
                str_content.append('{0} = {1}'.format(cn, cvi))

        str_content.append('')

    str_content = "\n".join(str_content)
    writeTextFile(file_name, str_content)


def mkDir(dir):
    try:
        _ext.os.makedirs(dir)
    except OSError:
        if not _ext.ospath.exists(dir):
            raise


def rmTree(dir, verbose=True):
    stat = _ext.stat

    if not _ext.ospath.exists(dir):
        return

    if verbose:
        _log.infoLabel('Removing: ', dir)

    chmodTree(dir, stat.S_IRWXU, stat.S_IRUSR | stat.S_IWUSR)

    if _ext.ospath.isdir(dir):
        _ext.shutil.rmtree(dir)
    else:
        _ext.os.unlink(dir)


def chmodTree(dir, dperm, fperm, keep_execute=False):
    os = _ext.os
    ospath = _ext.ospath
    stat = _ext.stat

    st_mode = os.lstat(dir).st_mode

    if stat.S_ISLNK(st_mode):
        lchmod(dir, fperm)
        return
    elif not stat.S_ISDIR(st_mode):
        os.chmod(dir, fperm)
        return

    walk_list = os.walk(dir)
    os.chmod(dir, dperm)

    for (path, dirs, files) in walk_list:
        for f in dirs + files:
            f = os.path.join(path, f)

            st_mode = os.lstat(f).st_mode

            if stat.S_ISLNK(st_mode):
                lchmod(f, fperm)
                continue

            if stat.S_ISDIR(st_mode):
                os.chmod(f, dperm)
            elif keep_execute:
                xperm = st_mode & (stat.S_IXUSR | stat.S_IXGRP)
                os.chmod(f, fperm | xperm)
            else:
                os.chmod(f, fperm)


def lchmod(target, perm):
    lchmod = getattr(_ext.os, 'lchmod', None)

    if lchmod:
        try:
            lchmod(target, perm)
        except OSError as e:
            # lchmod is falsely available in some Linux builds
            if e.errno != _ext.errno.ENOTSUP:
                raise


def cacheDir(key):
    os = _ext.os
    cache_dir = os.path.join(
        os.environ['HOME'], '.cache', 'futoin-cid', key)

    mkDir(cache_dir)

    return cache_dir


def tmpCacheDir(**kwargs):
    os = _ext.os
    stat = _ext.stat
    tmp_dir = cacheDir('tmp')

    # do once a day
    base_ts = int(_ext.time.time()) - (24 * 60 * 60)
    placeholder = os.path.join(tmp_dir, 'cleanup.stamp')

    if os.path.exists(placeholder) and os.stat(placeholder).st_mtime > base_ts:
        pass
    else:
        for f in os.listdir(tmp_dir):
            fp = os.path.join(tmp_dir, f)
            s = os.stat(fp)

            if s.st_mtime <= base_ts:
                if stat.S_ISDIR(s.st_mode):
                    _ext.pathutil.rmTree(fp)
                else:
                    os.remove(fp)
        _ext.pathutil.writeTextFile(placeholder, '')

    return _ext.tempfile.mkdtemp(dir=tmp_dir, **kwargs)


def downloadFile(env, url, dst):
    dst_tmp = dst + '.tmp'

    rmTree(dst_tmp)
    rmTree(dst)

    kwargs = {}
    _ext.configutil.requestsOptions(env, kwargs)
    kwargs['stream'] = True
    res = _ext.requests.request('GET', url, **kwargs)
    res.raise_for_status()

    with open(dst_tmp, 'wb') as f:
        _ext.shutil.copyfileobj(res.raw, f)

    _ext.os.rename(dst_tmp, dst)

def cacheDownloadFile(env, url):
    ospath = _ext.ospath
    url_hash = _ext.hashlib.sha256(url).hexdigest()
    cache_dir = _ext.pathutil.cacheDir('dl_{0}'.format(url_hash))
    cache_file = ospath.join(cache_dir, url.split('/')[-1])

    if not ospath.exists(cache_file):
        downloadFile(env, url, cache_file)

    return cache_file

def downloadStream(env, url, cmd):
    kwargs = {}
    _ext.configutil.requestsOptions(env, kwargs)
    kwargs['stream'] = True
    res = _ext.requests.request('GET', url, **kwargs)
    res.raise_for_status()

    _ext.executil.callExternal(cmd, input_stream=res.raw)


def downloadExtract(env, url, dst, tar_algo, strip=0):
    dst_tmp = dst + '.tmp'

    rmTree(dst_tmp)
    rmTree(dst)

    _ext.os.makedirs(dst_tmp)
    cmd = ['tar', 'x' + tar_algo, '-C', dst_tmp]

    if strip:
        cmd += ['--strip-components={0}'.format(strip)]

    downloadStream(env, url, cmd)
    _ext.os.rename(dst_tmp, dst)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .bundlermixin import BundlerMixIn


class pumaTool(BundlerMixIn, RuntimeTool):
    """A ruby web server built for concurrency http://puma.io

NOTE: the tool can only be used in scope of Ruby project through
bundler
"""
    __slots__ = ()

    def _sigReload(self):
        return self._ext.signal.SIGUSR2

    def getDeps(self):
        return ['bundler']

    def tuneDefaults(self, env):
        return {
            'minMemory': '128M',
            'connMemory': '8M',
            'connFD': 8,
            'socketTypes': ['unix', 'tcp'],
            'socketType': 'unix',
            'socketProtocol': 'http',
            'scalable': True,
            'reloadable': True,
            'multiCore': False,  # use workers on service level
            'maxRequestSize': '1M',
        }

    def onRun(self, config, svc, args):
        svc_tune = svc['tune']

        #---

        if svc_tune['socketType'] == 'unix':
            socket = 'unix://{0}'.format(svc_tune['socketPath'])
        else:
            socket = 'tcp://{0}:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])

        #---
        resource = self._ext.resource
        heap_limit = self._configutil.parseMemory(svc_tune['maxMemory'])
        # both limit RAM and HEAP (not the same)
        resource.setrlimit(resource.RLIMIT_RSS, (heap_limit, heap_limit))
        resource.setrlimit(resource.RLIMIT_DATA, (heap_limit, heap_limit))

        #---
        threads = svc_tune['maxConnections']

        puma_args = [
            '-b', socket,
            '-e', self._environ['RUBY_ENV'],
            '-t', '{0}:{0}'.format(threads),
            '--dir', self._os.getcwd(),  # force chdir() on reload
        ]

        #---
        env = config['env']

        cmd = [
            env['bundlerBin'],
            'exec', env['pumaBin'],
            svc['path']
        ] + puma_args + args

        self._ensureInstalled(env)
        self._executil.callInteractive(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import log as _log

_dev_null = None


def devNull():
    g = globals()
    ret = g['_dev_null']

    if ret is None:
        ret = open(_ext.os.devnull, 'w')
        g['_dev_null'] = ret

    return ret


def callMeaningful(cmd):
    callExternal(cmd, show_output=True)


def callExternal(cmd, suppress_fail=False, verbose=True,
                 output_handler=None, input=False,
                 merge_stderr=False, cwd=None,
                 user_interaction=False,
                 encoding='UTF-8',
                 binary_input=False, binary_output=False,
                 input_stream=None, show_output=False):

    sys = _ext.sys
    subprocess = _ext.subprocess

    if verbose and not suppress_fail:
        _log.infoLabel('Call: ', subprocess.list2cmdline(cmd))
        sys.stdout.flush()
        sys.stderr.flush()

    try:
        if input or input_stream:
            stdin = subprocess.PIPE
        elif user_interaction:
            stdin = None
        else:
            stdin = devNull()

        if merge_stderr:
            stderr = subprocess.STDOUT
        elif show_output:
            stderr = sys.stderr
        elif verbose and not suppress_fail:
            stderr = sys.stderr
        else:
            stderr = devNull()

        if user_interaction and not output_handler:
            stdout = None
        elif show_output:
            stdout = None
        else:
            stdout = subprocess.PIPE

        chunk_size = 65536
        res_buffers = []
        p = subprocess.Popen(cmd, stdin=stdin, stderr=stderr,
                             bufsize=chunk_size * 2, close_fds=True,
                             stdout=stdout, cwd=cwd)

        if input:
            if not binary_input:
                input = toBytes(input, encoding)

            try:
                p.stdin.write(input)
                p.stdin.flush()
            finally:
                p.stdin.close()

        if input_stream:
            _ext.shutil.copyfileobj(input_stream, p.stdin)
            p.stdin.flush()
            p.stdin.close()

        if stdout:
            if output_handler:
                on_chunk = output_handler
            elif binary_output:
                def on_binary_chunk(x):
                    res_buffers.append(x)

                on_chunk = on_binary_chunk
            else:
                def on_str_chunk(x):
                    res_buffers.append(toString(x, encoding))

                on_chunk = on_str_chunk

            try:
                while True:
                    chunk = p.stdout.read(chunk_size)

                    if chunk:
                        on_chunk(chunk)
                    else:
                        break
            finally:
                while p.stdout.read(chunk_size):
                    pass

        p.wait()

        if binary_output:
            res = b''.join(res_buffers)
        else:
            res = ''.join(res_buffers)

        if p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, cmd, res)

        return res
    except subprocess.CalledProcessError:
        if suppress_fail:
            return None
        raise


def callInteractive(cmd, replace=True, search_path=False, *args, **kwargs):
    if replace:
        if args or kwargs:
            _log.errorExit(
                'Extra args are not supported for replace call')

        sys = _ext.sys
        subprocess = _ext.subprocess
        os = _ext.os

        _log.infoLabel('Exec: ', subprocess.list2cmdline(cmd))
        sys.stdout.flush()
        sys.stderr.flush()

        # There is a problem of left FDs in Python 2
        #---
        import resource

        maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]

        if (maxfd == resource.RLIM_INFINITY):
            maxfd = 10240

        for fd in range(3, maxfd):
            try:
                os.close(fd)
            except OSError:
                pass
        #---

        if search_path:
            os.execvp(cmd[0], cmd)
        else:
            os.execv(cmd[0], cmd)
    else:
        callExternal(cmd, user_interaction=True, *args, **kwargs)


def trySudoCall(cmd, errmsg=None, **kwargs):
    try:
        if _ext.detect.isAdmin():
            callExternal(cmd, **kwargs)
            return

        environ = _ext.os.environ

        if environ.get('CID_INTERACTIVE_SUDO', '0') == '1':
            sudo_cmd = 'sudo -H'
            user_interaction = True
        else:
            sudo_cmd = 'sudo -n -H'
            user_interaction = False

        sudo_cmd = _ext.os.environ.get('cidSudo', sudo_cmd).split()
        callExternal(sudo_cmd + cmd,
                     user_interaction=user_interaction, **kwargs)
    except _ext.subprocess.CalledProcessError:
        if not errmsg:
            errmsg = 'you may need to call the the failed command manually !'

        _log.warn(errmsg)


def externalSetup(env, args):
    ext_cmd = env['externalSetup'].split()
    callExternal(ext_cmd + args)


def startService(name):
    openrc = '/sbin/rc-service'
    systemctl = '/bin/systemctl'
    ospath = _ext.ospath

    if ospath.exists(systemctl):
        trySudoCall(
            [systemctl, 'start', name],
            errmsg='you may need to start the service manually')
    elif ospath.exists(openrc):
        trySudoCall(
            [openrc, name, 'start'],
            errmsg='you may need to start the service manually')


def toString(value, encoding='UTF-8'):
    try:
        return str(value, encoding)
    except TypeError:
        return str(value)


def toBytes(value, encoding='UTF-8'):
    try:
        return value.encode(encoding=encoding)
    except AttributeError:
        return value
    except UnicodeDecodeError:
        return value



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class BashToolMixIn(object):
    __slots__ = ()

    def getDeps(self):
        return ['bash']

    def _callBash(self, env, cmd=None, bash_args=[], *args, **nargs):
        if cmd:
            cmd_args = ['-c', cmd]
        else:
            cmd_args = ['-s']

        return self._executil.callExternal(
            [env['bashBin'], '--noprofile', '--norc'] + cmd_args + bash_args,
            *args, **nargs)

    def _callBashInteractive(self, env, cmd, replace=True):
        return self._executil.callInteractive([
            env['bashBin'],
            '--noprofile', '--norc',
            '-c', cmd,
        ], replace=replace)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .ondemand import OnDemandMixIn


class DataSlots(OnDemandMixIn):
    __slots__ = (
        # ConfigMixIn
        '_startup_env',
        '_env',
        '_config',
        '_overrides',
        '_global_config',
        '_user_config',
        '_deploy_config',
        '_project_config',

        # DeployMixIn
        '_current_dir',
        '_devserve_mode',

        # LockMixIn
        '_deploy_lock',
        '_master_lock',
        '_global_lock',

        # ToolMixIn
        '_tool_impl',

        '_lastPackages',

        # ServiceMixIn
        '_running',
        '_reload_services',
        '_interruptable',
    )

    _FUTOIN_JSON = 'futoin.json'
    _FUTOIN_USER_JSON = '.futoin.json'
    _FUTOIN_MERGED_JSON = '.futoin.merged.json'



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..subtool import SubTool


class GemToolMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['gem', 'ruby']

    def _gemName(self):
        return self._name

    def _installTool(self, env):
        if self._detect.isExternalToolsSetup(env):
            self._executil.externalSetup(env, ['build-dep', 'ruby'])
        else:
            self._builddep.require(env, 'ruby')

        ver = env.get(self._name + 'Ver', None)
        version_arg = []

        if ver:
            version_arg = ['--version', ver]

        self._executil.callExternal([env['gemBin'], 'install', self._gemName(
        )] + env['gemInstallArgs'].split(' ') + version_arg)

    def installTool(self, env):
        if not self._have_tool:
            if self._detect.isDisabledToolsSetup(env):
                self._errorExit(
                    'Tool "{0}" must be installed externally (env config)'.format(self._name))
            else:
                self._warn(
                    'Auto-installing required "{0}" tool'.format(self._name))
                self._installTool(env)

            self.initEnv(env)

            if not self._have_tool:
                self._errorExit('Failed to install "{0}"'.format(self._name))

    def _updateTool(self, env):
        if env.get(self._name + 'Ver', None):
            self._installTool(env)
        else:
            self._executil.callExternal(
                [env['gemBin'], 'update', self._gemName()] + env['gemInstallArgs'].split(' '))

    def updateTool(self, env):
        if self._detect.isDisabledToolsSetup(env):
            self._errorExit(
                'Tool "{0}" must be updated externally (env config)'.format(self._name))
        else:
            self._updateTool(env)

    def uninstallTool(self, env):
        self._executil.callExternal([env['gemBin'], 'uninstall',
                                     '--force', self._gemName()])
        self._have_tool = False



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn


class phpbuildTool(BashToolMixIn, RunEnvTool):
    """Builds PHP so that multiple versions can be used side by side.

Home: https://github.com/php-build/php-build
"""
    __slots__ = ()

    PHPBUILD_LATEST = 'master'

    def getDeps(self):
        return ['bash', 'git']

    def _installTool(self, env):
        phpbuild_dir = env['phpbuildDir']
        phpbuild_git = env.get(
            'phpbuildGit', 'https://github.com/php-build/php-build.git')
        phpbuild_ver = env.get('phpbuildVer', self.PHPBUILD_LATEST)

        self._callBash(env,
                       'git clone {1} {0}; \
               cd {0} && git fetch && git reset --hard && git checkout {2}'
                       .format(phpbuild_dir, phpbuild_git, phpbuild_ver))

    def _updateTool(self, env):
        if env.get('phpBinOnly', False):
            return

        phpbuild_dir = env['phpbuildDir']
        phpbuild_ver = env.get('phpbuildVer', self.PHPBUILD_LATEST)

        self._callBash(env,
                       'cd {0} && git fetch && git reset --hard && git checkout {1} && git pull --rebase'
                       .format(phpbuild_dir, phpbuild_ver))

    def uninstallTool(self, env):
        phpbuild_dir = env['phpbuildDir']
        self._pathutil.rmTree(phpbuild_dir)
        self._have_tool = False

    def envNames(self):
        return ['phpbuildDir', 'phpbuildBin', 'phpbuildGit', 'phpbuildVer']

    def initEnv(self, env):
        ospath = self._ospath
        phpbuild_dir = ospath.join(self._environ['HOME'], '.phpbuild')
        phpbuild_dir = env.setdefault('phpbuildDir', phpbuild_dir)
        phpbuild_bin = env.setdefault(
            'phpbuildBin', ospath.join(phpbuild_dir, 'bin', 'php-build'))
        self._have_tool = ospath.exists(phpbuild_bin)

        # special case when environment required binaries only
        if env.get('phpBinOnly', False):
            self._have_tool = True



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .sdkmantoolmixin import SdkmanToolMixIn


class scalaTool(SdkmanToolMixIn, RuntimeTool):
    """The Scala Programming Language.

Home: https://www.scala-lang.org/

Installed via SDKMan!

Requires Java >= 8.
"""
    __slots__ = ()

    def _minJava(self):
        return '8'

    def tuneDefaults(self, env):
        return {
            'minMemory': '256M',
            'debugOverhead': '128M',
            'connMemory': '100K',
            'debugConnOverhead': '1M',
            'socketType': 'tcp',
            'scalable': False,
            'reloadable': False,
            'multiCore': True,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class binutilsTool(RunEnvTool):
    """GNU Binutils.

Home: https://www.gnu.org/software/binutils/

If binutilsDir is set it must point to bin folder.
"""
    __slots__ = ()

    def envNames(self):
        return ['binutilsDir', 'binutilsPrefix', 'binutilsPostfix']

    def _installTool(self, env):
        self._install.deb(['binutils'])
        self._install.rpm(['binutils'])
        self._install.emerge(['sys-devel/binutils'])
        self._install.pacman(['binutils'])
        self._install.apk(['binutils'])

    def initEnv(self, env):
        ospath = self._ospath
        pref = env.setdefault('binutilsPrefix', '')
        postf = env.setdefault('binutilsPostfix', '')

        bu_dir = env.get('binutilsDir', None)
        ld = pref + 'ld' + postf

        if bu_dir:
            ld = ospath.join(bu_dir, ld)
            self._have_tool = ospath.exists(bu_dir)
        else:
            ld = self._pathutil.which(ld)

            if ld:
                env['binutilsDir'] = ospath.dirname(ld)
                self._have_tool = True



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn


class nodeTool(BashToolMixIn, RuntimeTool):
    """Node.js is a JavaScript runtime built on Chrome's V8 JavaScript engine.

Home: https://nodejs.org/en/

Notes on tuning:
* .tune.nodeArgs

Note: NODE_ENV is set based on nodeEnv or .env.type
"""
    __slots__ = ()

    def getDeps(self):
        if self._detect.isAlpineLinux():
            return []

        return ['nvm', 'bash']

    def _installTool(self, env):
        if self._detect.isAlpineLinux():
            if env['nodeVer'] == 'current':
                self._install.apk('nodejs-current')
            else:
                self._install.apk('nodejs')
            return

        self._callBash(env,
                       'source {0} --no-use && nvm install {1}'
                       .format(env['nvmInit'], env['nodeVer']))

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        self._callBash(env,
                       'source {0} --no-use && nvm deactivate && nvm uninstall {1}'
                       .format(env['nvmInit'], env['nodeVer']))
        self._have_tool = False

    def envNames(self):
        return ['nodeBin', 'nodeVer', 'nodeEnv']

    def initEnv(self, env):
        node_version = env.setdefault('nodeVer', 'lts/*')
        #---
        node_env = env.get('nodeEnv', '')

        if node_env:
            pass
        elif env['type'] == 'dev':
            node_env = 'development'
        else:
            node_env = 'production'

        self._environ['NODE_ENV'] = node_env
        #---

        if self._detect.isAlpineLinux():
            super(nodeTool, self).initEnv(env)
            return

        try:
            env_to_set = self._callBash(env,
                                        'source {0} --no-use && \
                nvm use {1} >/dev/null && \
                env | egrep "^(NVM_|PATH=)"'
                                        .format(env['nvmInit'], node_version),
                                        verbose=False
                                        )
        except:
            return

        if env_to_set:
            self._pathutil.updateEnvFromOutput(env_to_set)
            super(nodeTool, self).initEnv(env)
            self._environ['nodeVer'] = node_version

    def tuneDefaults(self, env):
        return {
            'minMemory': '64M',
            'debugOverhead': '32M',
            'connMemory': '32K',
            'debugConnOverhead': '64K',
            'socketTypes': ['unix', 'tcp', 'tcp6'],
            'socketType': 'unix',
            'scalable': True,
            'reloadable': False,
            'multiCore': False,
            'maxRequestSize': '1M',
            'socketProtocol': 'http',
        }

    def onRun(self, config, svc, args):
        svc_tune = svc['tune']

        #---
        node_env = {}

        try:
            if svc_tune['socketType'] == 'unix':
                node_env['PORT'] = svc_tune['socketPath']
            else:
                node_env['PORT'] = str(svc_tune['socketPort'])
                node_env['HOST'] = svc_tune['socketAddress']
        except KeyError:
            pass

        self._environ.update(node_env)

        #---
        node_args = []

        if 'maxMemory' in svc_tune:
            heap_limit = self._configutil.parseMemory(svc_tune['maxMemory'])
            heap_limit = int(heap_limit * 0.9)
            heap_limit = int(heap_limit // 1024 // 1024)

            node_args.append(
                '--max_old_space_size={0}'.format(heap_limit)
            )

        #---

        cmd = [
            config['env']['nodeBin'],
            svc['path']
        ] + node_args + args

        self._executil.callInteractive(cmd)
